from io import StringIO

from utils import match

from chellow.models import User
from chellow.reports.report_csv_site_hh_data import content
from chellow.utils import ct_datetime, to_utc


def test_do_get(client, mocker, sess):
    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch(
        "chellow.reports.report_csv_site_hh_data.open_file", return_value=mock_file
    )
    MockThread = mocker.patch(
        "chellow.reports.report_csv_site_hh_data.threading.Thread"
    )
    user = User.get_by_email_address(sess, "admin@example.com")
    start_date = to_utc(ct_datetime(2020, 1, 1))
    finish_date = to_utc(ct_datetime(2020, 1, 31, 23, 30))
    site_id = None
    expected_args = start_date, finish_date, site_id, user.id
    query_string = {
        "start_year": "2020",
        "start_month": "01",
        "start_day": "01",
        "start_hour": "00",
        "start_minute": "00",
        "finish_year": "2020",
        "finish_month": "01",
        "finish_day": "31",
        "finish_hour": "23",
        "finish_minute": "30",
    }
    response = client.get("/reports/csv_site_hh_data", query_string=query_string)
    match(response, 303)
    MockThread.assert_called_with(target=content, args=expected_args)



from decimal import Decimal
from io import BytesIO
from zipfile import ZipFile

from openpyxl import Workbook

import pytest

from chellow.e.dno_rate_parser import (
    find_rates,
    str_to_hr,
    tab_llfs,
    tab_lv_hv,
    to_llfcs,
    to_pcs,
)


@pytest.mark.parametrize(
    "llfc_str,llfc_list",
    [
        ["H00-H01", ["H00", "H01"]],
        ["310 - 311 inclusive", ["310", "311"]],
        [
            """310
311
""",
            ["310", "311"],
        ],
    ],
)
def test_to_llfcs(mocker, llfc_str, llfc_list):
    actual = to_llfcs(llfc_str)
    assert actual == llfc_list


@pytest.mark.parametrize(
    "pc_str,pc_list", [["3 to 8 or 0", ["03", "04", "05", "06", "07", "08", "00"]]]
)
def test_to_pcs(mocker, pc_str, pc_list):
    mocker.patch("chellow.e.dno_rate_parser.get_value", return_value=pc_str)

    row = mocker.Mock()
    idx = mocker.Mock()
    actual = to_pcs(row, idx)
    assert actual == pc_list


@pytest.mark.parametrize("hr_str,hr", [["1600", Decimal("16")]])
def test_str_to_hr(mocker, hr_str, hr):
    actual = str_to_hr(hr_str)
    assert actual == hr


def test_find_rates():
    file_name = "a.zip"
    file_like = BytesIO()
    with ZipFile(file_like, "w") as zf:
        zf.writestr("b.txt", "")
    file_like.seek(0)
    find_rates(file_name, file_like)


def test_tab_llfs():
    wb = Workbook()
    wb.create_sheet("Annex 5 LLFs")
    sheet = wb.worksheets[1]
    sheet.insert_rows(0, 23)
    sheet.insert_cols(0, 9)
    sheet["A14"].value = "Metered voltage"
    sheet["A15"].value = "Low-voltage network"
    sheet["H15"].value = "100,ABB"

    vls = tab_llfs(sheet)
    assert vls == [
        {"code": "100", "voltage_level": "LV", "is_substation": False},
        {"code": "ABB", "voltage_level": "LV", "is_substation": False},
    ]


def test_tab_llfs_blank():
    wb = Workbook()
    wb.create_sheet("Annex 5 LLFs")
    sheet = wb.worksheets[1]
    sheet.insert_rows(0, 23)
    sheet.insert_cols(0, 9)
    sheet["A14"].value = "Metered voltage"
    sheet["A15"].value = "Low-voltage network"
    sheet["H15"].value = None

    vls = tab_llfs(sheet)
    assert vls == []


def test_tab_lv_hv():
    wb = Workbook()
    wb.create_sheet("Annex 1 LV, HV and UMS charges")
    sheet = wb.worksheets[1]
    sheet.insert_rows(0, 23)
    sheet.insert_cols(0, 12)
    sheet["A5"].value = "Time periods"
    sheet["A6"].value = "Saturday and Sunday\nAll year"
    sheet["B6"].value = ""
    sheet["C6"].value = "09:30 - 21:30"
    sheet["E6"].value = "00:00 - 09:30\n21:20 - 24:00"
    sheet["A7"].value = "Notes"

    rates = {}
    tab_lv_hv(sheet, rates)
    assert rates == {
        "bands": [
            {
                "band": "amber",
                "finish": Decimal("21.5"),
                "start": Decimal("9.5"),
                "weekend": True,
            }
        ],
        "tariffs": {},
    }



import csv
import sys
import threading
import traceback

from flask import g, request

from sqlalchemy import null, or_, true
from sqlalchemy.orm import joinedload

from chellow.dloads import open_file
from chellow.models import Era, Session, Site, SiteEra, Source, Supply, User
from chellow.utils import req_date, req_int, write_row
from chellow.views import chellow_redirect


TYPE_ORDER = {"hh": 0, "amr": 1, "nhh": 2, "unmetered": 3}


def content(start_date, finish_date, site_id, user_id):
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("site_hh_data.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            write_row(
                writer,
                "Site Id",
                "Site Name",
                "Associated Site Ids",
                "Sources",
                "Generator Types",
                "HH Start Clock-Time",
                "Imported kWh",
                "Displaced kWh",
                "Exported kWh",
                "Used kWh",
                "Parasitic kWh",
                "Generated kWh",
                "3rd Party Import",
                "3rd Party Export",
                "Meter Type",
            )

            sites = sess.query(Site).order_by(Site.code)
            if site_id is not None:
                sites = sites.filter(Site.id == site_id)

            for site in sites:
                sources = set()
                generator_types = set()
                assoc = site.find_linked_sites(sess, start_date, finish_date)
                metering_type = ""
                for era in (
                    sess.query(Era)
                    .join(SiteEra)
                    .filter(
                        SiteEra.site == site,
                        SiteEra.is_physical == true(),
                        Era.start_date <= finish_date,
                        or_(Era.finish_date == null(), Era.finish_date >= start_date),
                        Source.code != "sub",
                    )
                    .options(
                        joinedload(Era.supply).joinedload(Supply.source),
                        joinedload(Era.supply).joinedload(Supply.generator_type),
                    )
                ):
                    mtype = era.meter_category
                    if (
                        metering_type == ""
                        or TYPE_ORDER[mtype] < TYPE_ORDER[metering_type]
                    ):
                        metering_type = mtype
                    sources.add(era.supply.source.code)
                    generator_type = era.supply.generator_type
                    if generator_type is not None:
                        generator_types.add(generator_type.code)

                assoc_str = ",".join(sorted(s.code for s in assoc))
                sources_str = ",".join(sorted(list(sources)))
                generators_str = ",".join(sorted(list(generator_types)))

                for hh in site.hh_data(sess, start_date, finish_date):
                    write_row(
                        writer,
                        site.code,
                        site.name,
                        assoc_str,
                        sources_str,
                        generators_str,
                        hh["start_date"],
                        hh["imp_net"],
                        hh["displaced"],
                        hh["exp_net"],
                        hh["used"],
                        hh["exp_gen"],
                        hh["imp_gen"],
                        hh["imp_3p"],
                        hh["exp_3p"],
                        metering_type,
                    )

    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    start_date = req_date("start")
    finish_date = req_date("finish")
    if "site_id" in request.values:
        site_id = req_int("site_id")
    else:
        site_id = None

    args = start_date, finish_date, site_id, g.user.id
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from collections import defaultdict, namedtuple
from copy import deepcopy
from datetime import datetime as Datetime
from functools import lru_cache
from itertools import combinations, count
from math import log10
from types import MappingProxyType

from dateutil.relativedelta import relativedelta

from sqlalchemy import Float, cast, or_, select, text
from sqlalchemy.orm import aliased, joinedload
from sqlalchemy.sql.expression import false, null

from werkzeug.exceptions import BadRequest

from zish import dumps

import chellow.bank_holidays
import chellow.utils
from chellow.models import (
    Bill,
    BillType,
    Channel,
    ClockInterval,
    Contract,
    Era,
    HhDatum,
    Llfc,
    MeasurementRequirement,
    MtcParticipant,
    Party,
    RateScript,
    ReadType,
    RegisterRead,
    SiteEra,
    Source,
    Ssc,
    Supply,
    Tpr,
    hh_after,
    hh_before,
)
from chellow.utils import (
    HH,
    PropDict,
    YEAR,
    c_months_u,
    ct_datetime,
    ct_datetime_now,
    hh_format,
    hh_max,
    hh_min,
    hh_range,
    loads,
    to_ct,
    to_utc,
    utc_datetime_now,
)


cons_types = ["construction", "commissioning", "operation"]
lec_cats = list(
    (f"{v}-kwh", f"hist-{v}-kwh")
    for v in [
        "import-net",
        "export-net",
        "import-gen",
        "export-gen",
        "import-3rd-party",
        "export-3rd-party",
    ]
)


@lru_cache()
def get_times(start_date, finish_date, forecast_date):
    if start_date > finish_date:
        raise Exception(
            "The start date {hh_format(start_date)} is after the finish date "
            "{hh_format(finish_date)}."
        )
    hist_start = start_date
    hist_finish = finish_date
    years_back = 0
    while hist_finish > forecast_date:
        hist_start -= relativedelta(years=1)
        hist_finish -= relativedelta(years=1)
        years_back += 1

    return MappingProxyType(
        {
            "history-finish": hist_finish,
            "history-start": hist_start,
            "years-back": years_back,
        }
    )


def contract_func(caches, contract, func_name):
    try:
        ns = caches["computer"]["funcs"][contract.id]
    except KeyError:
        try:
            ccache = caches["computer"]
        except KeyError:
            ccache = caches["computer"] = {}

        try:
            contr_func_cache = ccache["funcs"]
        except KeyError:
            contr_func_cache = ccache["funcs"] = {}

        try:
            ns = contr_func_cache[contract.id]
        except KeyError:
            ns = {"db_id": contract.id, "properties": contract.make_properties()}
            exec(contract.charge_script, ns)
            contr_func_cache[contract.id] = ns

    return ns.get(func_name, None)


def non_core_rate(sess, caches, contract_id_or_name, date):
    return hh_rate(sess, caches, contract_id_or_name, date, market_role_code="Z")


def hh_rate(sess, caches, contract_id_or_name, date, market_role_code=None):
    try:
        return caches["computer"]["rates"][market_role_code][contract_id_or_name][date]
    except KeyError:
        try:
            ccache = caches["computer"]
        except KeyError:
            ccache = caches["computer"] = {}

        try:
            rss_cache = ccache["rates"]
        except KeyError:
            rss_cache = ccache["rates"] = {}

        try:
            mr_cache = rss_cache[market_role_code]
        except KeyError:
            mr_cache = rss_cache[market_role_code] = {}

        try:
            cont_cache = mr_cache[contract_id_or_name]
        except KeyError:
            cont_cache = mr_cache[contract_id_or_name] = {}

        try:
            return cont_cache[date]
        except KeyError:
            if market_role_code is None:
                contract = Contract.get_by_id(sess, contract_id_or_name)
            else:
                contract = Contract.get_by_role_code_name(
                    sess, market_role_code, contract_id_or_name
                )

            year_after = date + YEAR
            year_before = date - YEAR

            rs = (
                sess.query(RateScript)
                .filter(
                    RateScript.contract == contract,
                    RateScript.start_date <= date,
                    or_(
                        RateScript.finish_date == null(), RateScript.finish_date >= date
                    ),
                )
                .first()
            )

            if rs is None:
                rs = (
                    sess.query(RateScript)
                    .filter(RateScript.contract == contract)
                    .order_by(RateScript.start_date.desc())
                    .first()
                )
                if date < rs.start_date:
                    cstart = year_before
                    cfinish = min(year_after, rs.start_date - HH)
                else:
                    cstart = max(rs.finish_date + HH, year_before)
                    cfinish = year_after
            else:
                cstart = max(rs.start_date, year_before)
                if rs.finish_date is None:
                    cfinish = year_after
                else:
                    cfinish = min(rs.finish_date, year_after)

            market_role_code = rs.contract.market_role.code
            if market_role_code == "M":
                seg = "/e/mop_rate_scripts/"
            elif market_role_code == "C":
                seg = "/e/dc_rate_scripts/"
            elif market_role_code == "X":
                seg = "/e/supplier_rate_scripts/"
            elif market_role_code == "Z":
                seg = "/non_core_rate_scripts/"
            elif market_role_code == "R":
                seg = "/e/dno_rate_scripts/"
            else:
                raise Exception(
                    f"The market role code {market_role_code} isn't recognized."
                )

            vals = PropDict(
                f"the rate script {chellow.utils.url_root}{seg}{rs.id} ",
                loads(rs.script),
                [],
            )
            for dt in hh_range(caches, cstart, cfinish):
                if dt not in cont_cache:
                    cont_cache[dt] = vals

            return vals


def forecast_date():
    now = ct_datetime_now()
    return to_utc(ct_datetime(now.year, now.month, 1))


def displaced_era(
    sess,
    caches,
    site,
    start_date,
    finish_date,
    forecast_date,
    has_scenario_generation=False,
):
    start_date_ct = to_ct(start_date)
    finish_date_ct = to_ct(finish_date)
    if (start_date_ct.year, start_date_ct.month) != (
        finish_date_ct.year,
        finish_date_ct.month,
    ):
        raise BadRequest(
            "The start and end dates of a displaced period must be within the "
            "same month"
        )
    has_displaced = False
    eras = {}
    for site_era in (
        sess.query(SiteEra)
        .join(Era)
        .join(Supply)
        .join(Source)
        .filter(
            SiteEra.site == site,
            Era.start_date <= finish_date,
            or_(Era.finish_date == null(), Era.finish_date >= start_date),
        )
        .options(
            joinedload(SiteEra.era).joinedload(Era.supply).joinedload(Supply.source)
        )
    ):
        era = site_era.era
        source_code = era.supply.source.code
        if site_era.is_physical and (
            source_code in ("gen", "gen-net")
            or (
                source_code == "net"
                and sess.query(Channel)
                .filter(
                    Channel.era == era,
                    Channel.imp_related == false(),
                    Channel.channel_type == "ACTIVE",
                )
                .first()
                is not None
            )
        ):
            has_displaced = True

        if source_code in ("net", "gen-net") and era.imp_mpan_core is not None:
            eras[
                "_".join(
                    (
                        era.pc.code,
                        str(era.exp_mpan_core is None),
                        hh_format(era.start_date),
                        era.imp_mpan_core,
                    )
                )
            ] = era

    if (has_displaced or has_scenario_generation) and len(eras) > 0:
        era = eras[sorted(eras.keys())[0]]
    else:
        era = None
    return era


def _tpr_dict(sess, caches, tpr_code):
    try:
        return caches["computer"]["tprs"][tpr_code]
    except KeyError:
        try:
            ccache = caches["computer"]
        except KeyError:
            ccache = caches["computer"] = {}

        try:
            tpr_cache = ccache["tprs"]
        except KeyError:
            tpr_cache = ccache["tprs"] = {}

        tpr = Tpr.get_by_code(sess, tpr_code)
        days_of_week = dict([i, []] for i in range(7))
        tpr_dict = {
            "days-of-week": days_of_week,
            "is-gmt": tpr.is_gmt,
            "datum-cache": {},
        }
        if tpr.is_teleswitch:
            contract = Contract.find_supplier_by_name(sess, "teleswitch")
            if contract is None:
                cis = []
            else:
                tprs = hh_rate(sess, caches, contract.id, utc_datetime_now())["tprs"]
                try:
                    cis = tprs[tpr_code]
                except KeyError:
                    raise BadRequest(
                        f"Can't find the TPR {tpr_code} in the rate script of the "
                        f"'teleswitch' supplier contract."
                    )
        else:
            cis = [
                {
                    "day_of_week": ci.day_of_week,
                    "start_month": ci.start_month,
                    "start_day": ci.start_day,
                    "start_hour": ci.start_hour,
                    "start_minute": ci.start_minute,
                    "end_month": ci.end_month,
                    "end_day": ci.end_day,
                    "end_hour": ci.end_hour,
                    "end_minute": ci.end_minute,
                }
                for ci in sess.query(ClockInterval).filter(ClockInterval.tpr == tpr)
            ]

        for ci in cis:
            days_of_week[ci["day_of_week"] - 1].append(
                {
                    "start-month": ci["start_month"] * 100 + ci["start_day"],
                    "start-hour": ci["start_hour"] + ci["start_minute"] / 60,
                    "end-month": ci["end_month"] * 100 + ci["end_day"],
                    "end-hour": ci["end_hour"] + ci["end_minute"] / 60,
                }
            )

        tpr_cache[tpr_code] = tpr_dict
        return tpr_dict


def is_tpr(sess, caches, tpr_code, hh_date):
    try:
        return _tpr_dict(sess, caches, tpr_code)["datum-cache"][hh_date]
    except KeyError:
        tpr_dict = _tpr_dict(sess, caches, tpr_code)
        datum_cache = tpr_dict["datum-cache"]

        days_of_week = tpr_dict["days-of-week"]
        datum = datum_range(sess, caches, 0, hh_date, hh_date)[0]
        pref = "utc-" if tpr_dict["is-gmt"] else "ct-"

        decimal_hour = datum[pref + "decimal-hour"]
        fractional_month = datum[pref + "month"] * 100 + datum[pref + "day"]

        result = False

        for ci in days_of_week[datum[pref + "day-of-week"]]:
            if (
                (
                    ci["start-hour"] < ci["end-hour"]
                    and ci["start-hour"] <= decimal_hour < ci["end-hour"]
                )
                or (
                    ci["start-hour"] >= ci["end-hour"]
                    and (
                        ci["start-hour"] <= decimal_hour
                        or decimal_hour < ci["end-hour"]
                    )
                )
            ) and ci["start-month"] <= fractional_month <= ci["end-month"]:
                result = True
                break
        datum_cache[hh_date] = result
        return result


def cache_level(cache, key):
    try:
        return cache[key]
    except BaseException:
        cache[key] = new_cache = {}
        return new_cache


def cached_dno_contract(sess, caches, dno_code):
    try:
        return caches["computer"]["dno_contracts"][dno_code]
    except KeyError:
        computer_cache = cache_level(caches, "computer")
        cont_cache = cache_level(computer_cache, "dno_contracts")
        dno_contract = Contract.get_dno_by_name(sess, dno_code)
        cont_cache[dno_code] = dno_contract
        return dno_contract


def datum_range(sess, caches, years_back, start_date, finish_date):
    try:
        return caches["computer"]["datum_range"][years_back][start_date][finish_date]
    except KeyError:
        computer_cache = cache_level(caches, "computer")
        range_cache = cache_level(computer_cache, "datum_range")
        yb_range_cache = cache_level(range_cache, years_back)
        start_range_cache = cache_level(yb_range_cache, start_date)
        datum_cache = cache_level(computer_cache, "datum")
        yb_datum_cache = cache_level(datum_cache, years_back)

        datum_list = []
        try:
            bh_contract_id = caches["computer"]["bh_contract_id"]
        except KeyError:
            bh_cont = Contract.get_non_core_by_name(sess, "bank_holidays")
            bh_contract_id = caches["computer"]["bh_contract_id"] = bh_cont.id

        for dt in hh_range(caches, start_date, finish_date):
            try:
                datum = yb_datum_cache[dt]
            except KeyError:
                hist_date = dt - relativedelta(years=years_back)
                ct_dt = to_ct(dt)

                utc_is_month_end = (dt + HH).day == 1 and dt.day != 1
                ct_is_month_end = (ct_dt + HH).day == 1 and ct_dt.day != 1

                utc_decimal_hour = dt.hour + dt.minute / 60
                ct_decimal_hour = ct_dt.hour + ct_dt.minute / 60

                bhs = hh_rate(sess, caches, bh_contract_id, dt)["bank_holidays"]

                bank_holidays = [b[5:] for b in bhs]
                utc_is_bank_holiday = dt.strftime("%m-%d") in bank_holidays
                ct_is_bank_holiday = ct_dt.strftime("%m-%d") in bank_holidays

                datum = MappingProxyType(
                    {
                        "hist-start": hist_date,
                        "start-date": dt,
                        "ct-day": ct_dt.day,
                        "utc-month": dt.month,
                        "utc-day": dt.day,
                        "utc-decimal-hour": utc_decimal_hour,
                        "utc-year": dt.year,
                        "utc-hour": dt.hour,
                        "utc-minute": dt.minute,
                        "ct-year": ct_dt.year,
                        "ct-month": ct_dt.month,
                        "ct-decimal-hour": ct_decimal_hour,
                        "ct-day-of-week": ct_dt.weekday(),
                        "utc-day-of-week": dt.weekday(),
                        "utc-is-bank-holiday": utc_is_bank_holiday,
                        "ct-is-bank-holiday": ct_is_bank_holiday,
                        "utc-is-month-end": utc_is_month_end,
                        "ct-is-month-end": ct_is_month_end,
                        "status": "X",
                        "imp-msp-kvarh": 0,
                        "imp-msp-kvar": 0,
                        "exp-msp-kvarh": 0,
                        "exp-msp-kvar": 0,
                        "msp-kw": 0,
                        "msp-kwh": 0,
                        "hist-import-net-kvarh": 0,
                        "hist-export-net-kvarh": 0,
                        "anti-msp-kwh": 0,
                        "anti-msp-kw": 0,
                        "imp-msp-kvarh": 0,
                        "exp-msp-kvarh": 0,
                        "imp-msp-kvar": 0,
                        "exp-msp-kvar": 0,
                        "hist-imp-msp-kvarh": 0,
                        "hist-kwh": 0,
                    }
                )
                yb_datum_cache[dt] = datum
            datum_list.append(datum)
        datum_tuple = start_range_cache[finish_date] = tuple(datum_list)
        return datum_tuple


class DataSource:
    def __init__(
        self,
        sess,
        start_date,
        finish_date,
        forecast_date,
        caches,
        era_maps,
        deltas=None,
    ):
        self.full_channels = True
        self.sess = sess
        self.caches = caches
        self.forecast_date = forecast_date
        self.start_date = start_date
        self.finish_date = finish_date
        if self.start_date > self.finish_date:
            raise Exception(
                f"The start date {hh_format(start_date)} is after the finish date "
                f"{hh_format(finish_date)}."
            )
        self.deltas = deltas
        times = get_times(start_date, finish_date, forecast_date)
        self.years_back = times["years-back"]
        self.history_start = times["history-start"]
        self.history_finish = times["history-finish"]

        self.is_green = False
        self.supplier_bill = defaultdict(int, {"problem": ""})
        self.mop_bill = defaultdict(int, {"problem": ""})
        self.dc_bill = defaultdict(int, {"problem": ""})
        self.hh_data = []
        self.supplier_bill_hhs = {}
        self.mop_bill_hhs = {}
        self.dc_bill_hhs = {}

        self.era_maps = {} if era_maps is None else deepcopy(era_maps)
        era_map = {}
        for em_start, em in sorted(self.era_maps.items()):
            if em_start <= start_date:
                era_map = em

        self.era_map = PropDict("scenario properties", era_map)
        self.era_map_llfcs = self.era_map.get("llfcs", {})
        self.era_map_pcs = self.era_map.get("pcs", {})
        self.era_map_sscs = self.era_map.get("sscs", {})
        self.era_map_supplier_contracts = self.era_map.get("supplier_contracts", {})
        self.era_map_dc_contracts = self.era_map.get("dc_contracts", {})
        self.era_map_mop_contracts = self.era_map.get("mop_contracts", {})
        self.era_map_cops = self.era_map.get("cops", {})
        self.era_map_comms = self.era_map.get("comms", {})
        self.era_map_mpan_cores = self.era_map.get("mpan_cores", {})
        self.era_map_dnos = self.era_map.get("dnos", {})
        self.era_map_gsp_groups = self.era_map.get("gsp_groups", {})

    def contract_func(self, contract, func_name):
        return contract_func(self.caches, contract, func_name)

    def hh_rate(self, contract_id, date):
        try:
            return self.rate_cache[contract_id][date]
        except KeyError:
            return hh_rate(self.sess, self.caches, contract_id, date)
        except AttributeError:
            val = hh_rate(self.sess, self.caches, contract_id, date)
            self.rate_cache = self.caches["computer"]["rates"]
            return val

    def non_core_rate(self, contract_name_or_id, date):
        return non_core_rate(self.sess, self.caches, contract_name_or_id, date)

    def _add_problem(self, problem):
        self.supplier_bill["problem"] += problem
        self.mop_bill["problem"] += problem
        self.dc_bill["problem"] += problem


class SiteSource(DataSource):
    def __init__(
        self,
        sess,
        site,
        start_date,
        finish_date,
        forecast_date,
        caches,
        era=None,
        era_maps=None,
        deltas=None,
        bill=None,
        exclude_virtual=False,
    ):
        DataSource.__init__(
            self,
            sess,
            start_date,
            finish_date,
            forecast_date,
            caches,
            era_maps,
            deltas=deltas,
        )
        self.site = site
        self.bill = bill
        self.stream_focus = "gen-used"
        if era is None:
            self.is_displaced = False
        else:
            self.era = era
            self.is_displaced = True
            self.supply = self.era.supply
            self.mpan_core = era.imp_mpan_core
            self.dno = self.supply.dno
            self.dno_code = self.dno.dno_code
            self.dno_contract = cached_dno_contract(self.sess, caches, self.dno_code)

            era_map_llfcs = self.era_map_llfcs.get(self.dno_code, {})
            if era.imp_llfc.code in era_map_llfcs:
                self.llfc_code = era_map_llfcs[era.imp_llfc.code]
                self.llfc = self.dno.get_llfc_by_code(sess, self.llfc_code, start_date)
            else:
                self.llfc = era.imp_llfc
                self.llfc_code = self.llfc.code

            if era.pc.code in self.era_map_pcs:
                self.pc_code = self.era_map_pcs[era.pc.code]
            else:
                self.pc_code = era.pc.code

            era_ssc_code = None if era.ssc is None else era.ssc.code
            if era_ssc_code in self.era_map_sscs:
                self.ssc_code = self.era_map_sscs[era_ssc_code]
            else:
                self.ssc_code = era_ssc_code

            if era.cop.code in self.era_map_cops:
                self.cop_code = self.era_map_cops[era.cop.code]
            else:
                self.cop_code = era.cop.code

            if era.comm.code in self.era_map_comms:
                self.comm_code = self.era_map_comms[era.comm.code]
            else:
                self.comm_code = era.comm.code

            if era.imp_supplier_contract.id in self.era_map_supplier_contracts:
                self.supplier_contract = Contract.get_supplier_by_id(
                    sess, self.era_map_supplier_contracts[era.imp_supplier_contract.id]
                )
            else:
                self.supplier_contract = era.imp_supplier_contract

            if era.dc_contract.id in self.era_map_dc_contracts:
                self.dc_contract = Contract.get_dc_by_id(
                    sess, self.era_map_dc_contracts[era.dc_contract.id]
                )
            else:
                self.dc_contract = era.dc_contract

            if era.mop_contract.id in self.era_map_mop_contracts:
                self.mop_contract = Contract.get_mop_by_id(
                    sess, self.era_map_mop_contracts[era.mop_contract.id]
                )
            else:
                self.mop_contract = era.mop_contract

            if "sc" in self.era_map:
                self.sc = self.era_map["sc"]
            else:
                self.sc = era.imp_sc

            self.properties = self.era_map.get("properties_overwritten", {})
            self.properties.update(era.props)
            self.properties.update(self.era_map.get("properties_overwrite", {}))

            self.is_import = True
            self.voltage_level_code = self.llfc.voltage_level.code
            self.is_substation = self.llfc.is_substation
            self.gsp_group_code = self.supply.gsp_group.code
            self.ssc = self.era.ssc
            self.ssc_code = None if self.ssc is None else self.ssc.code
            self.energisation_status_code = era.energisation_status.code

        era_q = (
            select(Era.id)
            .join(SiteEra)
            .join(Supply)
            .join(Source)
            .where(
                SiteEra.site == site,
                Era.start_date <= self.history_finish,
                SiteEra.is_physical,
                Source.code != "sub",
                or_(Era.finish_date == null(), Era.finish_date >= self.history_start),
            )
        )
        if exclude_virtual:
            era_q = era_q.join(Party).where(Party.dno_code != "88")

        self.era_ids = set(sess.execute(era_q).scalars())

        if len(self.era_ids) == 0:
            rs = iter([])
        else:
            rs = iter(
                sess.query(
                    cast(HhDatum.value, Float),
                    HhDatum.start_date,
                    HhDatum.status,
                    Channel.imp_related,
                    Source.code,
                )
                .join(Channel)
                .join(Era)
                .join(Supply)
                .join(Source)
                .filter(
                    Channel.channel_type == "ACTIVE",
                    HhDatum.start_date >= self.history_start,
                    HhDatum.start_date <= self.history_finish,
                    Era.id.in_(list(self.era_ids)),
                )
                .order_by(HhDatum.start_date)
            )

        hh_value, hh_start_date, status, imp_related, source_code = next(
            rs, (None, None, None, None, None)
        )
        hist_map = {}

        for hist_date in hh_range(self.caches, self.history_start, self.history_finish):
            export_net_kwh = 0
            import_net_kwh = 0
            export_gen_kwh = 0
            import_gen_kwh = 0
            import_3rd_party_kwh = 0
            statuses = set()
            export_3rd_party_kwh = 0
            while hh_start_date == hist_date:
                statuses.add(status)
                if not imp_related and source_code in ("net", "gen-net"):
                    export_net_kwh += hh_value
                if imp_related and source_code in ("net", "gen-net"):
                    import_net_kwh += hh_value
                if (imp_related and source_code == "gen") or (
                    not imp_related and source_code == "gen-net"
                ):
                    import_gen_kwh += hh_value
                if (not imp_related and source_code == "gen") or (
                    imp_related and source_code == "gen-net"
                ):
                    export_gen_kwh += hh_value
                if (imp_related and source_code == "3rd-party") or (
                    not imp_related and source_code == "3rd-party-reverse"
                ):
                    import_3rd_party_kwh += hh_value
                if (not imp_related and source_code == "3rd-party") or (
                    imp_related and source_code == "3rd-party-reverse"
                ):
                    export_3rd_party_kwh += hh_value

                hh_value, hh_start_date, status, imp_related, source_code = next(
                    rs, (None, None, None, None, None)
                )

            if len(statuses) == 1 and statuses.pop() == "A":
                status = "A"
            else:
                status = "E"

            hh_values = {
                "status": status,
                "hist-import-net-kwh": import_net_kwh,
                "hist-export-net-kwh": export_net_kwh,
                "hist-import-gen-kwh": import_gen_kwh,
                "hist-export-gen-kwh": export_gen_kwh,
                "hist-import-3rd-party-kwh": import_3rd_party_kwh,
                "hist-export-3rd-party-kwh": export_3rd_party_kwh,
                "hist-used-3rd-party-kwh": import_3rd_party_kwh - export_3rd_party_kwh,
            }

            hh_values["used-3rd-party-kwh"] = hh_values["hist-used-3rd-party-kwh"]
            hh_values["hist-kwh"] = hh_values["hist-used-gen-msp-kwh"] = (
                hh_values["hist-import-gen-kwh"]
                - hh_values["hist-export-gen-kwh"]
                - hh_values["hist-export-net-kwh"]
            )

            hh_values["msp-kwh"] = hh_values["used-gen-msp-kwh"] = hh_values[
                "hist-used-gen-msp-kwh"
            ]
            for cat_kwh, hist_cat_kwh in lec_cats:
                hh_values[cat_kwh] = hh_values[hist_cat_kwh]

            hh_values["hist-used-kwh"] = (
                hh_values["hist-used-gen-msp-kwh"]
                + hh_values["hist-import-net-kwh"]
                + hh_values["hist-used-3rd-party-kwh"]
            )

            hh_values["used-kwh"] = hh_values["hist-used-kwh"]
            hh_values["import-net-kwh"] = hh_values["hist-import-net-kwh"]
            hh_values["msp-kw"] = hh_values["used-gen-msp-kw"] = (
                hh_values["used-gen-msp-kwh"] * 2
            )

            hist_map[hist_date] = hh_values

        for dtm in datum_range(
            sess, self.caches, self.years_back, start_date, finish_date
        ):
            datum = dtm.copy()
            datum.update(hist_map[datum["hist-start"]])
            datum["msp-kva"] = (
                datum["msp-kw"] ** 2
                + max(datum["imp-msp-kvar"], datum["exp-msp-kvar"]) ** 2
            ) ** 0.5
            self.hh_data.append(datum)
            self.supplier_bill_hhs[dtm["start-date"]] = {}
            self.mop_bill_hhs[dtm["start-date"]] = {}
            self.dc_bill_hhs[dtm["start-date"]] = {}

        if self.deltas is not None:
            for hh in self.hh_data:
                try:
                    delt_hh = self.deltas["hhs"][hh["start-date"]]
                    hh["import-net-kwh"] = delt_hh["import-net-kwh"]
                    hh["export-net-kwh"] = delt_hh["export-net-kwh"]
                    hh["import-gen-kwh"] = delt_hh["import-gen-kwh"]
                    hh["msp-kwh"] = delt_hh["msp-kwh"]
                    hh["used-kwh"] = delt_hh["used-kwh"]
                    hh["msp-kw"] = hh["msp-kwh"] * 2
                except KeyError:
                    pass

    def revolve_to_3rd_party_used(self):
        for hh in self.hh_data:
            hh["msp-kwh"] = hh["used-3rd-party-kwh"]
            hh["msp-kw"] = hh["msp-kwh"] * 2
            hh["hist-kwh"] = (
                hh["hist-import-3rd-party-kwh"] - hh["hist-export-3rd-party-kwh"]
            )
            hh["hist-kw"] = hh["hist-kwh"] * 2
        self.stream_focus = "3rd-party-used"

    def revolve_to_gen_used(self):
        for hh in self.hh_data:
            hh["msp-kwh"] = hh["used-gen-msp-kwh"]
            hh["msp-kw"] = hh["msp-kwh"] * 2
            hh["hist-kwh"] = hh["hist-used-gen-msp-kwh"]
            hh["hist-kw"] = hh["hist-kwh"] * 2
        self.stream_focus = "gen-used"

    def get_data_sources(self, start_date, finish_date, forecast_date=None):
        if forecast_date is None:
            forecast_date = self.forecast_date

        if all(
            (
                self.start_date == start_date,
                self.finish_date == finish_date,
                forecast_date == self.forecast_date,
            )
        ):
            yield self
        else:
            start_date_ct, finish_date_ct = to_ct(start_date), to_ct(finish_date)
            months = c_months_u(
                start_year=start_date_ct.year,
                start_month=start_date_ct.month,
                finish_year=finish_date_ct.year,
                finish_month=finish_date_ct.month,
            )
            for month_start, month_finish in months:
                chunk_start = hh_max(month_start, start_date)
                chunk_finish = hh_min(month_finish, finish_date)
                if self.stream_focus == "gen-used":
                    era = displaced_era(
                        self.sess,
                        self.caches,
                        self.site,
                        chunk_start,
                        chunk_finish,
                        forecast_date,
                    )
                    if era is None:
                        return
                else:
                    era = self.era

                site_ds = SiteSource(
                    self.sess,
                    self.site,
                    chunk_start,
                    chunk_finish,
                    forecast_date,
                    self.caches,
                    era,
                    self.era_maps,
                    self.deltas,
                )
                if self.stream_focus == "3rd-party-used":
                    site_ds.revolve_to_3rd_party_used()

                yield site_ds


ACTUAL_READ_TYPES = ["N", "N3", "C", "X", "CP"]


class SupplySource(DataSource):
    def __init__(
        self,
        sess,
        start_date,
        finish_date,
        forecast_date,
        era,
        is_import,
        caches,
        bill=None,
        era_maps=None,
        deltas=None,
    ):
        DataSource.__init__(
            self,
            sess,
            start_date,
            finish_date,
            forecast_date,
            caches,
            era_maps,
            deltas=deltas,
        )

        self.is_displaced = False
        self.bill = bill
        if self.bill is not None:
            self.bill_start = bill.start_date
            self.bill_finish = bill.finish_date
            self.is_last_bill_gen = (
                not self.bill_finish < self.start_date
                and not self.bill_finish > self.finish_date
            )

        self.site = None
        self.supply = era.supply
        self.supply_name = self.supply.name
        self.source_code = self.supply.source.code
        self.era = era

        if self.supply.generator_type is None:
            self.generator_type_code = None
        else:
            self.generator_type_code = self.supply.generator_type.code

        if era.supply.gsp_group.code in self.era_map_gsp_groups:
            self.gsp_group_code = self.era_map_gsp_groups[era.supply.gsp_group.code]
        else:
            self.gsp_group_code = era.supply.gsp_group.code

        if self.supply.dno.dno_code in self.era_map_dnos:
            self.dno_code = self.era_map_dnos[self.supply.dno.dno_code]
            self.dno = Party.get_dno_by_code(sess, self.dno_code, self.start_date)
        else:
            self.dno = self.supply.dno
            self.dno_code = self.dno.dno_code

        self.dno_contract = cached_dno_contract(self.sess, caches, self.dno_code)

        era_map_llfcs = self.era_map_llfcs.get(self.dno_code, {})
        self.is_import = is_import
        if is_import:
            if era.imp_mpan_core in self.era_map_mpan_cores:
                self.mpan_core = self.era_map_mpan_cores[era.imp_mpan_core]
            else:
                self.mpan_core = era.imp_mpan_core

            if era.imp_llfc.code in era_map_llfcs:
                llfc_code = era_map_llfcs[era.imp_llfc.code]
                self.llfc = self.dno.get_llfc_by_code(sess, llfc_code, start_date)
            else:
                self.llfc = era.imp_llfc

            self.sc = era.imp_sc
            self.supplier_account = era.imp_supplier_account

            if era.imp_supplier_contract.id in self.era_map_supplier_contracts:
                self.supplier_contract = Contract.get_supplier_by_id(
                    sess, self.era_map_supplier_contracts[era.imp_supplier_contract.id]
                )
            else:
                self.supplier_contract = era.imp_supplier_contract
        else:
            if era.exp_mpan_core in self.era_map_mpan_cores:
                self.mpan_core = self.era_map_mpan_cores[era.exp_mpan_core]
            else:
                self.mpan_core = era.exp_mpan_core

            if era.exp_llfc is None:
                self.mpan_core = llfc_key = "new_export"
                llfc_key = "new_export"
            else:
                self.mpan_core = era.exp_mpan_core
                llfc_key = era.exp_llfc.code
            if llfc_key in era_map_llfcs:
                llfc_code = era_map_llfcs[llfc_key]
                self.llfc = self.dno.get_llfc_by_code(sess, llfc_code, start_date)
            else:
                self.llfc = era.exp_llfc

            self.sc = 0 if era.exp_sc is None else era.exp_sc
            self.supplier_account = era.exp_supplier_account

            if era.exp_supplier_contract is None:
                sup_key = "new_export"
            else:
                sup_key = era.exp_supplier_contract.id

            if sup_key in self.era_map_supplier_contracts:
                self.supplier_contract = Contract.get_supplier_by_id(
                    sess, self.era_map_supplier_contracts[sup_key]
                )
            else:
                self.supplier_contract = era.exp_supplier_contract

        if era.dc_contract.id in self.era_map_dc_contracts:
            self.dc_contract = Contract.get_dc_by_id(
                sess, self.era_map_dc_contracts[era.dc_contract.id]
            )
        else:
            self.dc_contract = era.dc_contract

        if era.mop_contract.id in self.era_map_mop_contracts:
            self.mop_contract = Contract.get_mop_by_id(
                sess, self.era_map_mop_contracts[era.mop_contract.id]
            )
        else:
            self.mop_contract = era.mop_contract

        if era.cop.code in self.era_map_cops:
            self.cop_code = self.era_map_cops[era.cop.code]
        else:
            self.cop_code = era.cop.code

        if era.comm.code in self.era_map_comms:
            self.comm_code = self.era_map_comms[era.comm.code]
        else:
            self.comm_code = era.comm.code

        self.properties = dict(self.era_map.get("properties_overwritten", {}))
        self.properties.update(era.props)
        self.properties.update(self.era_map.get("properties_overwrite", {}))

        self.id = self.mpan_core
        self.msn = era.msn
        self.llfc_code = self.llfc.code
        self.voltage_level = self.llfc.voltage_level
        self.voltage_level_code = self.voltage_level.code
        self.is_substation = self.llfc.is_substation
        self.is_new = False
        self.mtc_participant = self.era.mtc_participant
        self.meter_type = self.mtc_participant.meter_type
        self.meter_type_code = self.meter_type.code
        self.mtc = self.mtc_participant.mtc

        if era.pc.code in self.era_map_pcs:
            self.pc_code = self.era_map_pcs[era.pc.code]
        else:
            self.pc_code = era.pc.code

        era_ssc_code = None if era.ssc is None else era.ssc.code
        if era_ssc_code in self.era_map_sscs:
            ssc_code = self.era_map_sscs[era_ssc_code]
            if ssc_code is None:
                self.ssc = None
            else:
                self.ssc = Ssc.get_by_code(sess, ssc_code)
        else:
            self.ssc = era.ssc

        self.ssc_code = None if self.ssc is None else self.ssc.code
        self.energisation_status_code = self.era.energisation_status.code

        self.measurement_type = era.meter_category

        self.consumption_info = ""
        self.normal_reads = set()
        hist_map = {}

        if self.years_back == 0:
            hist_eras = [self.era]
        else:
            hist_eras = (
                sess.query(Era)
                .filter(
                    Era.supply == self.supply,
                    Era.start_date <= self.history_finish,
                    or_(
                        Era.finish_date == null(), Era.finish_date >= self.history_start
                    ),
                )
                .order_by(Era.start_date)
            )
            if self.is_import:
                hist_eras = hist_eras.filter(Era.imp_mpan_core != null())
            else:
                hist_eras = hist_eras.filter(Era.exp_mpan_core != null())
            hist_eras = hist_eras.options(
                joinedload(Era.pc),
                joinedload(Era.channels),
                joinedload(Era.mtc_participant).joinedload(MtcParticipant.meter_type),
            ).all()
            if len(hist_eras) == 0:
                hist_eras = (
                    sess.query(Era)
                    .filter(Era.supply == self.supply)
                    .order_by(Era.start_date)
                )
                if self.is_import:
                    hist_eras = hist_eras.filter(Era.imp_mpan_core != null())
                else:
                    hist_eras = hist_eras.filter(Era.exp_mpan_core != null())
                hist_eras = hist_eras.limit(1).all()

        for i, hist_era in enumerate(hist_eras):
            if self.is_import:
                hist_mpan_core = hist_era.imp_mpan_core
            else:
                hist_mpan_core = hist_era.exp_mpan_core

            if hist_mpan_core is None:
                continue

            if self.history_start > hist_era.start_date:
                chunk_start = self.history_start
            else:
                if i == 0:
                    chunk_start = self.history_start
                else:
                    chunk_start = hist_era.start_date

            chunk_finish = hh_min(self.history_finish, hist_era.finish_date)

            hist_measurement_type = hist_era.meter_category

            if hist_measurement_type == "amr" and self.era_map.get(
                "use_amr_hh_data", False
            ):
                hist_measurement_type = "hh"

            if hist_measurement_type == "unmetered":
                kwh = (
                    hist_era.imp_sc
                    * 60
                    * 30
                    / (
                        Datetime(chunk_start.year + 1, 1, 1)
                        - Datetime(chunk_start.year, 1, 1)
                    ).total_seconds()
                )

                for tpr in (
                    sess.query(Tpr)
                    .join(MeasurementRequirement)
                    .filter(MeasurementRequirement.ssc == hist_era.ssc)
                ):
                    for hist_date in hh_range(self.caches, chunk_start, chunk_finish):
                        if is_tpr(sess, self.caches, tpr.code, hist_date):
                            hist_map[hist_date] = {
                                "msp-kw": kwh * 2,
                                "msp-kwh": kwh,
                                "hist-kwh": kwh,
                                "imp-msp-kvarh": 0,
                                "imp-msp-kvar": 0,
                                "exp-msp-kvarh": 0,
                                "exp-msp-kvar": 0,
                                "status": "A",
                                "tpr": tpr.code,
                            }

            elif self.bill is None and hist_measurement_type in ("nhh", "amr"):
                if self.energisation_status_code == "E":
                    pairs, self.normal_reads = _no_bill_nhh(
                        sess,
                        caches,
                        self.supply,
                        chunk_start,
                        chunk_finish,
                        hist_map,
                        forecast_date,
                    )
                    self.consumption_info += dumps(pairs)

            elif self.bill is not None and hist_measurement_type in ("nhh", "amr"):
                if self.energisation_status_code == "E":
                    hhd = {}
                    for hh_date in hh_range(self.caches, chunk_start, chunk_finish):
                        hhd[hh_date] = {
                            "status": "X",
                            "imp-msp-kvarh": 0,
                            "imp-msp-kvar": 0,
                            "exp-msp-kvarh": 0,
                            "exp-msp-kvar": 0,
                            "msp-kw": 0,
                            "msp-kwh": 0,
                            "hist-kwh": 0,
                        }

                    tpr_codes = [
                        t[0]
                        for t in sess.query(Tpr.code)
                        .join(MeasurementRequirement)
                        .filter(MeasurementRequirement.ssc == self.ssc)
                    ]

                    bills = dict(
                        (b.id, b)
                        for b in sess.query(Bill)
                        .filter(
                            Bill.supply == self.supply,
                            Bill.start_date <= chunk_finish,
                            Bill.finish_date >= chunk_start,
                        )
                        .order_by(Bill.issue_date.desc(), Bill.start_date)
                    )
                    while True:
                        to_del = None
                        for a, b in combinations(bills.values(), 2):
                            if all(
                                (
                                    a.start_date == b.start_date,
                                    a.finish_date == b.finish_date,
                                    a.net == -1 * b.net,
                                    a.vat == -1 * b.vat,
                                    a.gross == -1 * b.gross,
                                )
                            ):
                                to_del = (a.id, b.id)
                                break
                        if to_del is None:
                            break
                        else:
                            for k in to_del:
                                del bills[k]

                    prev_type_alias = aliased(ReadType)
                    pres_type_alias = aliased(ReadType)
                    for bill in bills.values():
                        kws = defaultdict(int)
                        for (
                            coefficient,
                            previous_date,
                            previous_value,
                            previous_type,
                            present_date,
                            present_value,
                            present_type,
                            tpr_code,
                        ) in (
                            sess.query(
                                cast(RegisterRead.coefficient, Float),
                                RegisterRead.previous_date,
                                cast(RegisterRead.previous_value, Float),
                                prev_type_alias.code,
                                RegisterRead.present_date,
                                cast(RegisterRead.present_value, Float),
                                pres_type_alias.code,
                                Tpr.code,
                            )
                            .join(Tpr)
                            .join(
                                prev_type_alias,
                                RegisterRead.previous_type_id == prev_type_alias.id,
                            )
                            .join(
                                pres_type_alias,
                                RegisterRead.present_type_id == pres_type_alias.id,
                            )
                            .filter(RegisterRead.bill == bill, RegisterRead.units == 0)
                            .order_by(RegisterRead.present_date)
                        ):
                            if tpr_code not in tpr_codes:
                                self._add_problem(
                                    f"The TPR {tpr_code} from the register read does "
                                    f"not match any of the TPRs "
                                    f"({', '.join(tpr_codes)}) associated with the "
                                    f"MPAN."
                                )

                            if present_date > bill.finish_date:
                                self._add_problem(
                                    "There's a read after the end of the bill!"
                                )
                            advance = present_value - previous_value
                            if advance < 0:
                                self._add_problem("Clocked?")
                                digits = int(log10(previous_value)) + 1
                                advance = 10**digits - previous_value + present_value

                            kwh = advance * coefficient
                            self.consumption_info += (
                                f"dumb nhh kwh for {tpr_code} from "
                                f"{hh_format(chunk_start)} to "
                                f"{hh_format(chunk_finish)} is {kwh}\n"
                            )

                            kws[tpr_code] += kwh

                        for tpr_code, kwh in kws.items():
                            if (
                                present_type in ACTUAL_READ_TYPES
                                and previous_type in ACTUAL_READ_TYPES
                            ):
                                status = "A"
                            else:
                                status = "E"

                            hh_part = {}
                            for hh_date in hh_range(
                                self.caches, bill.start_date, bill.finish_date
                            ):
                                if is_tpr(sess, self.caches, tpr_code, hh_date):
                                    hh_part[hh_date] = {"status": status}

                            num_hh = len(hh_part)
                            if num_hh > 0:
                                rate = kwh / num_hh
                                for d, h in hh_part.items():
                                    if chunk_start <= d <= chunk_finish:
                                        hhd_datum = hhd[d]
                                        hhd_datum["msp-kw"] += rate * 2
                                        hhd_datum["msp-kwh"] += rate
                                        hhd_datum["hist-kwh"] += rate
                                        hhd_datum["tpr"] = tpr_code
                                        if hhd_datum["status"] in ("X", "A"):
                                            hhd_datum["status"] = h["status"]
                            elif kwh > 0:
                                self._add_problem(
                                    f"For the TPR code {tpr_code} the bill says "
                                    f"that there are {kwh} kWh, but the time of "
                                    f"the TPR doesn't cover the time between the "
                                    f"register reads."
                                )

                    hist_map.update(hhd)
            elif hist_measurement_type == "hh":
                full_channels, hhd = _init_hh_data(
                    sess, caches, hist_era, chunk_start, chunk_finish, is_import
                )
                if not full_channels:
                    self.full_channels = False
                hist_map.update(hhd)

            else:
                raise BadRequest("gen type not recognized")

        for d in datum_range(
            sess, self.caches, self.years_back, start_date, finish_date
        ):
            datum = d.copy()
            datum.update(hist_map.get(d["hist-start"], {}))
            datum["msp-kva"] = (
                datum["msp-kw"] ** 2
                + max(datum["imp-msp-kvar"], datum["exp-msp-kvar"]) ** 2
            ) ** 0.5
            self.hh_data.append(datum)

            d_start = d["start-date"]
            self.supplier_bill_hhs[d_start] = {}
            self.mop_bill_hhs[d_start] = {}
            self.dc_bill_hhs[d_start] = {}

        if self.deltas is not None:
            site_deltas = self.deltas["site"]

            try:
                sup_deltas = self.deltas[self.supply.id]
            except KeyError:
                sup_deltas = self.deltas[self.supply.id] = {}

            for hh in self.hh_data:
                hh_start = hh["start-date"]
                if hh_start in sup_deltas:
                    delt = sup_deltas[hh_start]
                elif hh_start in site_deltas:
                    delt = sup_deltas[hh_start] = site_deltas[hh_start]
                    del site_deltas[hh_start]
                else:
                    continue

                hh["msp-kwh"] += delt
                hh["msp-kw"] += delt * 2

    def get_data_sources(self, start_date, finish_date, forecast_date=None):
        if forecast_date is None:
            forecast_date = self.forecast_date

        if all(
            (
                self.start_date == start_date,
                self.finish_date == finish_date,
                forecast_date == self.forecast_date,
            )
        ):
            yield self
        elif self.mpan_core in ("new_export", "new_import"):
            yield SupplySource(
                self.sess,
                start_date,
                finish_date,
                forecast_date,
                self.era,
                self.is_import,
                self.caches,
                self.bill,
                self.era_maps,
                self.deltas,
            )
        else:
            if self.is_import:
                eras = self.sess.query(Era).filter(
                    Era.supply == self.supply,
                    Era.imp_mpan_core != null(),
                    Era.start_date <= finish_date,
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                )
            else:
                eras = self.sess.query(Era).filter(
                    Era.supply == self.supply,
                    Era.exp_mpan_core != null(),
                    Era.start_date <= finish_date,
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                )
            eras = eras.options(
                joinedload(Era.ssc),
                joinedload(Era.dc_contract),
                joinedload(Era.mop_contract),
                joinedload(Era.imp_supplier_contract),
                joinedload(Era.exp_supplier_contract),
                joinedload(Era.channels),
                joinedload(Era.imp_llfc).joinedload(Llfc.voltage_level),
                joinedload(Era.exp_llfc).joinedload(Llfc.voltage_level),
                joinedload(Era.cop),
                joinedload(Era.comm),
                joinedload(Era.supply).joinedload(Supply.dno),
                joinedload(Era.supply).joinedload(Supply.gsp_group),
                joinedload(Era.supply).joinedload(Supply.source),
                joinedload(Era.mtc_participant).joinedload(MtcParticipant.meter_type),
                joinedload(Era.pc),
                joinedload(Era.site_eras),
            )

            for era in eras:
                chunk_start = hh_max(era.start_date, start_date)
                chunk_finish = hh_min(era.finish_date, finish_date)

                ds = SupplySource(
                    self.sess,
                    chunk_start,
                    chunk_finish,
                    forecast_date,
                    era,
                    self.is_import,
                    self.caches,
                    self.bill,
                    self.era_maps,
                    self.deltas,
                )
                yield ds


Read = namedtuple("Read", "date msn type reads")


def _find_pair(sess, caches, is_forwards, read_list):
    if len(read_list) < 2:
        return

    if is_forwards:
        back, front = read_list[-2], read_list[-1]
    else:
        back, front = read_list[-1], read_list[-2]

    back_reads = back["reads"]
    front_reads = front["reads"]
    back_date = back["date"]
    front_date = front["date"]

    if back["msn"] == front["msn"] and back_reads.keys() == front_reads.keys():
        r = hh_range(caches, back_date, front_date - HH)

        tprs = {}
        for tpr_code, initial_val in back_reads.items():
            end_val = front_reads[tpr_code]
            num_hh = len([d for d in r if is_tpr(sess, caches, tpr_code, d)])

            # Clocked?
            if end_val - initial_val < 0:
                digits = int(log10(initial_val)) + 1
                end_val += 10**digits

            kwh = (
                end_val * front["coefficients"][tpr_code]
                - initial_val * back["coefficients"][tpr_code]
            )

            tprs[tpr_code] = kwh / num_hh if num_hh > 0 else 0

        pair_reads = set()
        for r in back, front:
            desc = tuple([f"{k}: {v}" for k, v in r["reads"].items()])
            pair_reads.add(Read(r["date"], r["msn"], r["read_type"], desc))

        return {
            "reads": pair_reads,
            "start-date": back_date,
            "tprs": tprs,
        }


def _find_hhs(caches, sess, pairs, chunk_start, chunk_finish):
    if len(pairs) == 0:
        pairs.append({"start-date": chunk_start, "tprs": {"00001": 0}})

    # set finish dates
    for i in range(1, len(pairs)):
        pairs[i - 1]["finish-date"] = pairs[i]["start-date"] - HH
    pairs[-1]["finish-date"] = None

    # stretch
    if hh_after(pairs[0]["start-date"], chunk_start):
        pairs[0]["start-date"] = chunk_start

    # chop
    if hh_before(pairs[0]["finish-date"], chunk_start):
        del pairs[0]
    if hh_after(pairs[-1]["start-date"], chunk_finish):
        del pairs[-1]

    # squash
    if hh_before(pairs[0]["start-date"], chunk_start):
        pairs[0]["start-date"] = chunk_start
    if hh_after(pairs[-1]["finish-date"], chunk_finish):
        pairs[-1]["finish-date"] = chunk_finish

    hhs = {}
    for pair in pairs:
        d_range = hh_range(caches, pair["start-date"], pair["finish-date"])
        for tpr_code, kwh in pair["tprs"].items():
            dates = [d for d in d_range if is_tpr(sess, caches, tpr_code, d)]

            for date in dates:
                hhs[date] = {
                    "msp-kw": kwh * 2,
                    "msp-kwh": kwh,
                    "hist-kwh": kwh,
                    "imp-msp-kvar": 0,
                    "imp-msp-kvarh": 0,
                    "exp-msp-kvar": 0,
                    "exp-msp-kvarh": 0,
                    "tpr": tpr_code,
                }
    return hhs


def _set_status(hhs, read_list, forecast_date):
    rl = [r["date"] for r in read_list if r["date"] <= forecast_date]
    last_read_date = rl[-1] if len(rl) > 0 else to_utc(ct_datetime(1970, 1, 1))
    for k, v in hhs.items():
        v["status"] = "E" if k > last_read_date else "A"


def _read_generator(sess, supply, start, is_forwards, is_prev):
    if is_prev:
        r_typ = RegisterRead.previous_type
        r_dt = RegisterRead.previous_date
        r_vl = RegisterRead.previous_value
    else:
        r_typ = RegisterRead.present_type
        r_dt = RegisterRead.present_date
        r_vl = RegisterRead.present_value

    q = (
        sess.query(RegisterRead)
        .join(Bill)
        .join(BillType)
        .join(r_typ)
        .filter(
            RegisterRead.units == 0,
            ReadType.code.in_(ACTUAL_READ_TYPES),
            Bill.supply == supply,
            BillType.code != "W",
        )
        .options(joinedload(RegisterRead.bill), joinedload(r_typ))
    )

    if is_forwards:
        q = q.filter(r_dt >= start).order_by(r_dt, RegisterRead.id)
    else:
        q = q.filter(r_dt < start).order_by(r_dt.desc(), RegisterRead.id)

    for offset in count():
        r = q.offset(offset).first()
        if r is None:
            break

        if is_prev:
            dt = r.previous_date
            read_type = r.previous_type
        else:
            dt = r.present_date
            read_type = r.present_type

        bill = (
            sess.query(Bill)
            .join(BillType)
            .filter(
                Bill.supply == supply,
                Bill.reads.any(),
                Bill.finish_date >= r.bill.start_date,
                Bill.start_date <= r.bill.finish_date,
                BillType.code != "W",
            )
            .order_by(Bill.issue_date.desc(), BillType.code)
            .first()
        )

        if bill.id != r.bill.id:
            continue

        era = supply.find_era_at(sess, dt)
        if era is None:
            era_coeff = None
        else:
            era_properties = PropDict(
                chellow.utils.url_root + "eras/" + str(era.id), loads(era.properties)
            )
            try:
                era_coeff = float(era_properties["coefficient"])
            except KeyError:
                era_coeff = None

        reads = {}
        coeffs = {}
        for coeff, value, tpr_code in sess.query(
            cast(RegisterRead.coefficient, Float), cast(r_vl, Float), Tpr.code
        ).filter(
            RegisterRead.units == 0,
            RegisterRead.bill == r.bill,
            RegisterRead.msn == r.msn,
            RegisterRead.tpr_id == Tpr.id,
            r_dt == dt,
        ):
            reads[tpr_code] = value
            coeffs[tpr_code] = coeff if era_coeff is None else era_coeff

        yield {
            "date": dt,
            "reads": reads,
            "coefficients": coeffs,
            "msn": r.msn,
            "read_type": read_type.code,
        }


def _no_bill_nhh(sess, caches, supply, start, finish, hist_map, forecast_date):
    read_list = []
    pairs = []
    read_keys = set()
    normal_reads = set()

    for is_forwards in (False, True):
        prev_reads = iter(_read_generator(sess, supply, start, is_forwards, True))
        pres_reads = iter(_read_generator(sess, supply, start, is_forwards, False))
        if is_forwards:
            read_list.reverse()

        for read in _make_reads(is_forwards, prev_reads, pres_reads):
            read_key = read["date"], read["msn"]
            if read_key in read_keys:
                continue
            read_keys.add(read_key)

            read_list.append(read)
            pair = _find_pair(sess, caches, is_forwards, read_list)
            if pair is not None:
                pairs.append(pair)
                normal_reads = normal_reads | pair["reads"]
                if not is_forwards or (is_forwards and read_list[-1]["date"] > finish):
                    break

    hhs = _find_hhs(caches, sess, pairs, start, finish)
    _set_status(hhs, read_list, forecast_date)
    hist_map.update(hhs)
    return pairs, normal_reads


def _make_reads(forwards, prev_reads, pres_reads):
    prev_read = next(prev_reads, None)
    pres_read = next(pres_reads, None)

    while prev_read is not None or pres_read is not None:
        if prev_read is None:
            yield pres_read
            pres_read = next(pres_reads, None)

        elif pres_read is None:
            yield prev_read
            prev_read = next(prev_reads, None)

        else:
            if (forwards and prev_read["date"] < pres_read["date"]) or (
                not forwards and prev_read["date"] >= pres_read["date"]
            ):
                yield prev_read
                prev_read = next(prev_reads, None)
            else:
                yield pres_read
                pres_read = next(pres_reads, None)


def _init_hh_data(sess, caches, hist_era, chunk_start, chunk_finish, is_import):
    hhd = {}
    full_channels = True
    has_exp_active = False
    has_imp_related_reactive = False
    has_exp_related_reactive = False
    for channel in hist_era.channels:
        if not channel.imp_related and channel.channel_type == "ACTIVE":
            has_exp_active = True
        if channel.imp_related and channel.channel_type in (
            "REACTIVE_IMP",
            "REACTIVE_EXP",
        ):
            has_imp_related_reactive = True
        if not channel.imp_related and channel.channel_type in (
            "REACTIVE_IMP",
            "REACTIVE_EXP",
        ):
            has_exp_related_reactive = True

    if has_exp_active and not has_exp_related_reactive and has_imp_related_reactive:
        #  old style
        full_channels = False
        data = iter(
            sess.execute(
                text(
                    """
select sum(cast(coalesce(kwh.value, 0) as double precision)),
sum(cast(coalesce(anti_kwh.value, 0) as double precision)),
max(kwh.status),
sum(cast(coalesce(reactive_imp.value, 0) as double precision)),
sum(cast(coalesce(reactive_exp.value, 0) as double precision)),
hh_datum.start_date
from hh_datum
join channel on (hh_datum.channel_id = channel.id)
left join hh_datum as kwh
on (hh_datum.id = kwh.id and channel.channel_type = 'ACTIVE'
and channel.imp_related = :is_import)
left join hh_datum as anti_kwh
on (hh_datum.id = anti_kwh.id and channel.channel_type = 'ACTIVE'
and channel.imp_related != :is_import)
left join hh_datum as reactive_imp
on (hh_datum.id = reactive_imp.id
and channel.channel_type = 'REACTIVE_IMP'
and channel.imp_related is true)
left join hh_datum as reactive_exp
on (hh_datum.id = reactive_exp.id
and channel.channel_type = 'REACTIVE_EXP'
and channel.imp_related is true)
where channel.era_id = :era_id and hh_datum.start_date >= :start_date
and hh_datum.start_date <= :finish_date
group by hh_datum.start_date
order by hh_datum.start_date
"""
                ),
                params={
                    "era_id": hist_era.id,
                    "start_date": chunk_start,
                    "finish_date": chunk_finish,
                    "is_import": is_import,
                },
            )
        )
        (msp_kwh, anti_msp_kwh, status, imp_kvarh, exp_kvarh, hist_start) = next(
            data, (None, None, None, None, None, None)
        )

        for hh_date in hh_range(caches, chunk_start, chunk_finish):
            if hh_date == hist_start:
                hhd[hh_date] = {
                    "status": status,
                    "imp-msp-kvarh": imp_kvarh,
                    "imp-msp-kvar": imp_kvarh * 2,
                    "exp-msp-kvarh": exp_kvarh,
                    "exp-msp-kvar": exp_kvarh * 2,
                    "msp-kw": msp_kwh * 2,
                    "msp-kwh": msp_kwh,
                    "anti-msp-kwh": anti_msp_kwh,
                    "hist-kwh": msp_kwh,
                }
                (
                    msp_kwh,
                    anti_msp_kwh,
                    status,
                    imp_kvarh,
                    exp_kvarh,
                    hist_start,
                ) = next(data, (None, None, None, None, None, None))
    else:
        # new style
        data = iter(
            sess.execute(
                text(
                    """
select hh_datum.start_date,
max(kwh.status),
sum(cast(coalesce(kwh.value, 0) as double precision)),
sum(cast(coalesce(reactive_imp.value, 0) as double precision)),
sum(cast(coalesce(reactive_exp.value, 0) as double precision))
from hh_datum
join channel on (hh_datum.channel_id = channel.id)
left join hh_datum as kwh
on (hh_datum.id = kwh.id and channel.channel_type = 'ACTIVE'
and channel.imp_related = :is_import)
left join hh_datum as reactive_imp
on (hh_datum.id = reactive_imp.id
and channel.channel_type = 'REACTIVE_IMP'
and channel.imp_related = :is_import)
left join hh_datum as reactive_exp
on (hh_datum.id = reactive_exp.id
and channel.channel_type = 'REACTIVE_EXP'
and channel.imp_related = :is_import)
where channel.era_id = :era_id and hh_datum.start_date >= :start_date
and hh_datum.start_date <= :finish_date
group by hh_datum.start_date
order by hh_datum.start_date
"""
                ),
                params={
                    "era_id": hist_era.id,
                    "start_date": chunk_start,
                    "finish_date": chunk_finish,
                    "is_import": is_import,
                },
            )
        )

        for hist_start, status, msp_kwh, imp_kvarh, exp_kvarh in data:
            datum = {
                "imp-msp-kvarh": imp_kvarh,
                "imp-msp-kvar": imp_kvarh * 2,
                "exp-msp-kvarh": exp_kvarh,
                "exp-msp-kvar": exp_kvarh * 2,
            }
            if msp_kwh is not None:
                datum["status"] = status
                datum["hist-kwh"] = msp_kwh
                datum["msp-kwh"] = msp_kwh
                datum["msp-kw"] = msp_kwh * 2

            hhd[hist_start] = datum
    return full_channels, hhd



from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest


from chellow.utils import parse_mpan_core, to_utc


def get_ct_date(row, idx):
    cell = get_cell(row, idx)
    val = cell.value
    if not isinstance(val, Datetime):
        raise BadRequest(f"Problem reading {val} as a timestamp at {cell.coordinate}.")
    return val


def get_start_date(row, idx):
    return to_utc(get_ct_date(row, idx))


def get_finish_date(row, idx):
    return to_utc(get_ct_date(row, idx) + relativedelta(hours=23, minutes=30))


def get_cell(row, idx):
    try:
        return row[idx]
    except IndexError:
        raise BadRequest(
            f"For the row {row}, the index is {idx} which is beyond the end of the row."
        )


def get_str(row, idx):
    return get_cell(row, idx).value.strip()


def get_dec(row, idx):
    cell = get_cell(row, idx)
    try:
        return Decimal(str(cell.value))
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_int(row, idx):
    return int(get_cell(row, idx).value)


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)
        self.sheet = self.book.worksheets[1]

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        try:
            bills = []
            row = next(self.sheet.iter_rows(min_row=6, max_row=6, max_col=3))
            issue_date = get_start_date(row, 2)
            for row in self.sheet.iter_rows(min_row=12, max_col=11):
                val = get_cell(row, 1).value
                if val is None or val == "":
                    break

                self._set_last_line(row[0].row, val)
                mpan_core = parse_mpan_core(str(get_int(row, 1)))
                comms = get_str(row, 2)

                settled_str = get_str(row, 3)
                if settled_str == "Settled":
                    settlement_status = "settlement"
                else:
                    settlement_status = "non_settlement"

                start_date = get_start_date(row, 5)
                finish_date = get_finish_date(row, 6)

                meter_rate = get_dec(row, 7)
                net = round(get_dec(row, 8), 2)
                vat = round(get_dec(row, 9), 2)
                gross = round(get_dec(row, 10), 2)

                breakdown = {
                    "raw-lines": [],
                    "comms": comms,
                    "settlement-status": [settlement_status],
                    "meter-rate": [meter_rate],
                    "meter-gbp": net,
                }
                bills.append(
                    {
                        "bill_type_code": "N",
                        "kwh": Decimal(0),
                        "net": net,
                        "vat": vat,
                        "gross": gross,
                        "reads": [],
                        "breakdown": breakdown,
                        "account": mpan_core,
                        "issue_date": issue_date,
                        "start_date": start_date,
                        "finish_date": finish_date,
                        "mpan_core": mpan_core,
                        "reference": "_".join(
                            (
                                start_date.strftime("%Y%m%d"),
                                finish_date.strftime("%Y%m%d"),
                                issue_date.strftime("%Y%m%d"),
                                mpan_core,
                            )
                        ),
                    }
                )
        except BadRequest as e:
            raise BadRequest(f"Row number {row[0].row}: {e.description}")

        return bills



import csv
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation
from itertools import count

from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook
from werkzeug.exceptions import BadRequest


from chellow.utils import HH, ct_datetime, parse_mpan_core, to_ct, to_utc


ELEM_MAP = {
    None: None,
    "Charge - Recurring": {
        None: ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days")
    },
    "Meter - Standard": {
        "Energy Bill Relief Scheme": {None: ["ebrs-gbp", "ebrs-msp-kwh"]},
        "Energy Bill Relief Scheme Discount": {None: ["ebrs-gbp", "ebrs-msp-kwh"]},
        "Energy Bill Discount Scheme": {None: ["ebrs-gbp", "ebrs-msp-kwh"]},
    },
    "Meter - UK Electricity - AAHEDC Pass-Thru": {
        None: ["aahedc-gbp", "aahedc-rate", "aahedc-gsp-kwh"]
    },
    "Meter - UK Electricity - BSUoS Pass-Thru": {
        None: ["bsuos-gbp", "bsuos-rate", "bsuos-nbp-kwh"]
    },
    "Meter - UK Electricity - Capacity Market Pass-Thru": {
        None: ["capacity-gbp", "capacity-rate", "capacity-kwh"],
        "Reverse Capacity Market Estimate": {None: ["capacity-gbp"]},
    },
    "Meter - UK Electricity - CfD FiT Pass-Thru": {
        None: ["cfd-fit-gbp", "cfd-fit-rate", "cfd-fit-nbp-kwh"]
    },
    "Meter - UK Electricity - CCL": {
        None: ["ccl-gbp", "ccl-rate", "ccl-kwh"],
        "CCL": {None: ["ccl-gbp", "ccl-rate", "ccl-kwh"]},
        "Levy Exempt Energy": {None: ["lec-gbp", "lec-rate", "lec-kwh"]},
    },
    "Meter - UK Electricity - DUoS": {
        None: None,
        "DUoS Unit Rate 3": {
            None: ("duos-green-gbp", "duos-green-rate", "duos-green-kwh")
        },
        "DUoS Unit Charge 3": {
            None: ("duos-green-gbp", "duos-green-rate", "duos-green-kwh")
        },
        "DUoS Unit Charge 2": {
            None: ("duos-amber-gbp", "duos-amber-rate", "duos-amber-kwh")
        },
        "DUoS Unit Rate 2": {
            None: ("duos-amber-gbp", "duos-amber-rate", "duos-amber-kwh")
        },
        "DUoS Unit Charge 1": {None: ("duos-red-gbp", "duos-red-kwh", "duos-red-rate")},
        "DUoS Unit Rate 1": {None: ("duos-red-gbp", "duos-red-kwh", "duos-red-rate")},
        "DUoS Standing Charge": {
            None: ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days")
        },
        "DUoS Fixed": {None: ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days")},
        "DUoS Reactive": {
            None: ("duos-reactive-gbp", "duos-reactive-rate", "duos-reactive-kvarh")
        },
    },
    "Meter - UK Electricity - FiT Pass-Thru": {
        None: ("fit-gbp", "fit-rate", "fit-msp-kwh")
    },
    "Pass Thru - UK Electricity Cost Component": {
        None: ("meter-rental-gbp", "meter-rental-rate", "meter-rental-days")
    },
    "Meter - UK Electricity - RO Pass-Thru": {
        None: ("ro-gbp", "ro-rate", "ro-msp-kwh")
    },
    "Meter - UK Electricity - TUoS": {
        None: ("triad-gbp", "triad-rate", "triad-gsp-kw"),
        "TNUoS Fixed": {
            None: ("tnuos-gbp", "tnuos-rate", "tnuos-days"),
        },
    },
    "Meter - UK Electricity - Standard": {
        None: None,
        "Unit Rate": {
            "Summer Weekday": {
                None: (
                    "summer-weekday-gbp",
                    "summer-weekday-rate",
                    "summer-weekday-gsp-kwh",
                )
            },
            "Peak": {None: ("peak-gbp", "peak-rate", "peak-gsp-kwh")},
            "Peak Shoulder": {
                None: (
                    "peak-shoulder-gbp",
                    "peak-shoulder-gsp-kwh",
                    "peak-shoulder-rate",
                )
            },
            "Summer Night": {
                None: ("summer-night-gbp", "summer-night-rate", "summer-night-gsp-kwh")
            },
            "Summer Weekend & Bank Holiday": {
                None: (
                    "summer-weekend-gbp",
                    "summer-weekend-rate",
                    "summer-weekend-gsp-kwh",
                )
            },
            "Night": {None: ("night-gbp", "night-rate", "night-gsp-kwh")},
            "Winter Weekday": {
                None: (
                    "winter-weekday-gbp",
                    "winter-weekday-rate",
                    "winter-weekday-gsp-kwh",
                )
            },
            "Winter Weekend & Bank Holiday": {
                None: (
                    "winter-weekend-gbp",
                    "winter-weekend-rate",
                    "winter-weekend-gsp-kwh",
                )
            },
            "Winter Night": {
                None: ("winter-night-gbp", "winter-night-rate", "winter-night-gsp-kwh")
            },
            "Day": {None: ("day-gbp", "day-rate", "day-gsp-kwh")},
            "Single": {None: ("day-gbp", "day-rate", "day-gsp-kwh")},
            "Off Peak / Weekends": {None: ("day-gbp", "day-rate", "day-gsp-kwh")},
        },
        "Reverse BSUoS in Unit Rate": {
            None: ("bsuos-reverse-gbp", "bsuos-reverse-rate", "bsuos-reverse-nbp-kwh")
        },
    },
    "Meter - UK Gas - CCL": {None: ("ccl-gbp", "ccl-rate", "ccl-kwh")},
}


def _find_names(tree, path):
    if len(path) > 0:
        try:
            return _find_names(tree[path[0]], path[1:])
        except KeyError:
            pass

    return tree[None]


COLUMNS = [
    "Billing Entity",
    "Customer Name",
    "Customer Number",
    "Account Name",
    "Account Number",
    "Billing Address",
    "Bill Number",
    "Bill Date",
    "Due Date",
    "Accepted Date",
    "Bill Period",
    "Agreement Number",
    "Product Bundle",
    "Product Name",
    "Bill Status",
    "Product Item Class",
    "Type",
    "Description",
    "From Date",
    "To Date",
    "Sales Tax Rate",
    "Meter Point",
    "Usage",
    "Usage Unit",
    "Price",
    "Amount",
    "Currency",
    "Indicator",
    "Product Item Name",
    "Rate Name",
]
COLUMN_MAP = dict(zip(COLUMNS, count(1)))


def get_date_naive(sheet, row, col):
    value = get_value(sheet, row, col)
    if value in ("", None):
        return None
    elif not isinstance(value, Datetime):
        raise BadRequest(
            f"Problem reading {value} as a timestamp at row {row} col {col}."
        )
    return value


def get_date(sheet, row, col_name):
    dt = get_date_naive(sheet, row, col_name)
    return None if dt is None else to_utc(to_ct(dt))


def get_value(sheet, row, col_name):
    col = COLUMN_MAP[col_name]
    try:
        return sheet.cell(row=row, column=col).value
    except IndexError:
        raise BadRequest(
            f"Can't find the cell at row {row} and col {col} on sheet {sheet}."
        )


def get_str(sheet, row, col):
    value = get_value(sheet, row, col)
    if value is None:
        return None

    return value.strip()


def get_dec(sheet, row, col):
    value = get_value(sheet, row, col)
    if value in ("", None):
        return None

    try:
        return Decimal(str(value))
    except InvalidOperation as e:
        raise BadRequest(
            f"Problem parsing the number '{value}' at row {row} col {col}. {e}"
        )


def get_int(sheet, row, col):
    value = get_value(sheet, row, col)
    try:
        return int(value)
    except ValueError as e:
        raise BadRequest(
            f"Problem parsing the integer '{value}' at row {row} col {col}. {e}"
        )


def _bd_add(bd, el_name, val):
    if el_name.split("-")[-1] in ("rate", "kva"):
        if el_name not in bd:
            bd[el_name] = set()
        bd[el_name].add(val)
    else:
        if el_name not in bd:
            bd[el_name] = 0
        try:
            bd[el_name] += val
        except TypeError as e:
            raise BadRequest(
                f"Problem with element name {el_name} and value '{val}': {e}"
            )


def _customer_mods(cust_number, description, bill):
    print(cust_number, type(cust_number), description)
    if cust_number == 10001596:
        if (
            description == "RO Mutualisation 2021-22"
            and "ro-gbp" in bill["breakdown"]
            and bill["issue_date"] == to_utc(ct_datetime(2023, 4, 14))
            and bill["start_date"] == to_utc(ct_datetime(2023, 3, 1))
            and bill["finish_date"] == to_utc(ct_datetime(2023, 3, 31, 23, 30))
        ):
            bill["start_date"] = to_utc(ct_datetime(2021, 4, 1))
            bill["finish_date"] = to_utc(ct_datetime(2022, 3, 31, 23, 30))

    return bill


def _parse_row(sheet, row, title_row):
    val = get_value(sheet, row, "Meter Point")
    try:
        mpan_core = parse_mpan_core(str(int(val)))
    except ValueError as e:
        raise BadRequest(
            f"Can't parse the MPAN core in column 'Meter Point' with value '{val}' : "
            f"{e}"
        )

    bill_period = get_value(sheet, row, "Bill Period")
    if "-" in bill_period:
        period_start_naive, period_finish_naive = [
            Datetime.strptime(v, "%Y-%m-%d") for v in bill_period.split(" - ")
        ]
        period_start = to_utc(to_ct(period_start_naive))
        period_finish = to_utc(to_ct(period_finish_naive + relativedelta(days=1) - HH))
    else:
        period_start, period_finish = None, None

    from_date = get_date(sheet, row, "From Date")
    if from_date is None:
        if period_start is None:
            raise BadRequest("Can't find a bill start date.")
        else:
            from_date = period_start

    to_date_naive = get_date_naive(sheet, row, "To Date")
    if to_date_naive is None:
        if period_finish is None:
            raise BadRequest("Can't find a bill finish date.")
        else:
            to_date = period_finish

    else:
        to_date = to_utc(to_ct(to_date_naive + relativedelta(days=1) - HH))

    issue_date = get_date(sheet, row, "Bill Date")
    bill_number = get_value(sheet, row, "Bill Number")
    bill = {
        "bill_type_code": "N",
        "kwh": Decimal(0),
        "vat": Decimal("0.00"),
        "net": Decimal("0.00"),
        "reads": [],
        "breakdown": {},
        "account": mpan_core,
        "issue_date": issue_date,
        "start_date": from_date,
        "finish_date": to_date,
        "mpan_core": mpan_core,
    }
    bd = bill["breakdown"]

    usage = get_dec(sheet, row, "Usage")
    # usage_units = get_value(sheet, row, 'Usage Unit')
    price = get_dec(sheet, row, "Price")
    amount = get_dec(sheet, row, "Amount")
    product_item_name = get_value(sheet, row, "Product Item Name")
    rate_name = get_value(sheet, row, "Rate Name")
    if product_item_name == "Renewables Obligation (RO)" and usage is not None:
        bill["kwh"] += round(usage, 2)
    description = get_value(sheet, row, "Description")
    product_class = get_value(sheet, row, "Product Item Class")
    if description in ("Standard VAT@20%", "Reduced VAT@5%"):
        bill["vat"] += round(amount, 2)
        if description.endswith("20%"):
            vat_percentage = Decimal("20")
        else:
            vat_percentage = Decimal("5")
        bd["vat_percentage"] = vat_percentage
    else:
        bill["net"] += round(amount, 2)

        path = ["" if v is None else v for v in [product_class, description, rate_name]]
        names = _find_names(ELEM_MAP, path)

        duos_avail_prefix = "DUoS Availability ("
        duos_excess_avail_prefix = "DUoS Excess Availability ("

        if description.startswith("DUoS Availability Adjustment "):
            _bd_add(bd, "duos-availability-gbp", amount)
        elif description.startswith("DUoS Availability"):
            if description.startswith(duos_avail_prefix):
                _bd_add(
                    bd,
                    "duos-availability-kva",
                    int(description[len(duos_avail_prefix) : -5]),
                )
            _bd_add(bd, "duos-availability-days", usage)
            _bd_add(bd, "duos-availability-rate", price)
            _bd_add(bd, "duos-availability-gbp", amount)
        elif description.startswith("DUoS Excess Availability"):
            if description.startswith(duos_excess_avail_prefix):
                kva = int(description[len(duos_excess_avail_prefix) : -5])
                _bd_add(bd, "duos-excess-availability-kva", kva)
            _bd_add(bd, "duos-excess-availability-days", usage)
            _bd_add(bd, "duos-excess-availability-rate", price)
            _bd_add(bd, "duos-excess-availability-gbp", amount)
        elif description.startswith("BSUoS Black Start "):
            _bd_add(bd, "black-start-gbp", amount)
        elif description.startswith("BSUoS Reconciliation - "):
            if usage is not None:
                _bd_add(bd, "bsuos-nbp-kwh", usage)
            if price is not None:
                _bd_add(bd, "bsuos-rate", price)
            _bd_add(bd, "bsuos-gbp", amount)
        elif description.startswith("FiT Rec - "):
            _bd_add(bd, "fit-gbp", amount)
        elif description.startswith("FiT Reconciliation "):
            _bd_add(bd, "fit-gbp", amount)
        elif description.startswith("CfD FiT Rec - ") or description.startswith(
            "CfD FiT Reconciliation"
        ):
            _bd_add(bd, "cfd-fit-gbp", amount)
        elif description.startswith("Flex"):
            _bd_add(bd, "reconciliation-gbp", amount)
        elif description.startswith("Legacy TNUoS Reversal "):
            _bd_add(bd, "triad-gbp", amount)
        elif description.startswith("Hand Held Read -"):
            _bd_add(bd, "meter-rental-gbp", amount)
        elif description.startswith("RO Mutualisation "):
            _bd_add(bd, "ro-gbp", amount)
        elif description.startswith("OOC MOP - "):
            _bd_add(bd, "meter-rental-gbp", amount)
        elif description.startswith("KVa Adjustment "):
            _bd_add(bd, "duos-availability-gbp", amount)
        elif names is not None:
            for elem_k, elem_v in zip(names, (amount, price, usage)):
                if elem_k is not None:
                    _bd_add(bd, elem_k, elem_v)
        else:
            raise BadRequest(
                f"For the path {path} the parser can't work out the element."
            )

    reference = f"{bill_number}_{row}"
    for k, v in tuple(bd.items()):
        if isinstance(v, set):
            bd[k] = list(v)
        elif k.endswith("-gbp"):
            reference += "_" + k[:-4]

    bill["reference"] = reference
    bill["gross"] = bill["net"] + bill["vat"]
    customer_number = get_value(sheet, row, "Customer Number")
    return _customer_mods(customer_number, description, bill)


def _make_raw_bills(sheet):
    bills = []
    rows = tuple(sheet.rows)
    title_row = rows[0]
    for row_index, row in enumerate(rows[1:], start=2):
        if len(row) > 21:
            val = row[21].value
            if val not in (None, ""):
                try:
                    bills.append(_parse_row(sheet, row_index, title_row))
                except BadRequest as e:
                    raise BadRequest(f"On row {row_index}: {e.description}")

    return bills


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)
        self.sheet = self.book.worksheets[0]

        self.last_line = None
        lines = (self._set_last_line(i, l) for i, l in enumerate(f))
        self.reader = csv.reader(lines, skipinitialspace=True)
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        return _make_raw_bills(self.sheet)



from utils import match

import chellow.dloads
from chellow.models import (
    Comm,
    Contract,
    Cop,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    User,
    UserRole,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_291 import content
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_http(mocker, sess, client):
    mocker.patch("chellow.reports.report_291.threading.Thread")
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    Contract.insert_dno(
        sess,
        dno.dno_code,
        participant,
        "",
        {},
        valid_from,
        None,
        {},
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(sess, "845", True, False, valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc_imp = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    dno.insert_llfc(
        sess,
        "521",
        "Export (HV)",
        voltage_level,
        False,
        False,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc_imp, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    sess.commit()

    query_string = {
        "supply_id": supply.id,
        "start_year": 2000,
        "start_month": 1,
        "start_day": 1,
        "start_hour": 0,
        "start_minute": 0,
        "finish_year": 2000,
        "finish_month": 1,
        "finish_day": 1,
        "finish_hour": 0,
        "finish_minute": 0,
    }

    response = client.get("/reports/291", query_string=query_string)

    match(response, 303)


def test_content(mocker, sess):
    mocker.patch("chellow.reports.report_291.threading.Thread")
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", True, False, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc_imp = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    dno.insert_llfc(sess, "521", "Export (HV)", voltage_level, False, False, vf, None)
    MtcLlfc.insert(sess, mtc_participant, llfc_imp, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    sess.commit()

    supply_id = supply.id
    start_date = to_utc(ct_datetime(2000, 1, 1))
    finish_date = to_utc(ct_datetime(2000, 1, 1))
    content(supply_id, start_date, finish_date, user_id)

    files = list(p.name for p in chellow.dloads.download_path.iterdir())
    assert files == ["00000_FINISHED_adminexamplecom_supply_virtual_bills_1.csv"]



import csv
import threading
import traceback

from flask import g, request

from sqlalchemy import or_
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

import chellow.e.computer
import chellow.e.duos
import chellow.e.tnuos
from chellow.dloads import open_file
from chellow.models import Era, Pc, Session, Site, SiteEra, Source, Supply, User
from chellow.utils import (
    HH,
    c_months_u,
    csv_make_val,
    ct_datetime,
    reduce_bill_hhs,
    req_int,
    to_utc,
)
from chellow.views import chellow_redirect


def _make_sites(sess, year_start, year_finish, site_id, source_codes):
    sites = (
        sess.query(Site)
        .join(SiteEra)
        .join(Era)
        .join(Pc)
        .join(Supply)
        .join(Source)
        .filter(
            Pc.code == "00",
            Source.code.in_(source_codes),
            Era.start_date <= year_finish,
            or_(Era.finish_date == null(), Era.finish_date >= year_start),
        )
        .distinct()
        .order_by(Site.code)
    )

    if site_id is not None:
        sites = sites.filter(Site.id == site_id)

    return sites


def _write_sites(sess, caches, writer, year, site_id):
    titles = (
        "Site Code",
        "Site Name",
        "Displaced TRIAD 1 Date",
        "Displaced TRIAD 1 MSP kW",
        "Displaced TRIAD LAF",
        "Displaced TRIAD 1 GSP kW",
        "Displaced TRIAD 2 Date",
        "Displaced TRIAD 2 MSP kW",
        "Displaced TRIAD 2 LAF",
        "Displaced TRIAD 2 GSP kW",
        "Displaced TRIAD 3 Date",
        "Displaced TRIAD 3 MSP kW",
        "Displaced TRIAD 3 LAF",
        "Displaced TRIAD 3 GSP kW",
        "Displaced GSP kW",
        "Displaced Rate GBP / kW",
        "GBP",
    )
    writer.writerow(titles)

    march_finish_ct = ct_datetime(year, 4, 1) - HH
    march_finish_utc = to_utc(march_finish_ct)
    march_start_ct = ct_datetime(year, 3, 1)
    march_start_utc = to_utc(march_start_ct)
    year_start = to_utc(ct_datetime(year - 1, 4, 1))

    forecast_date = chellow.e.computer.forecast_date()

    sites = _make_sites(sess, year_start, march_finish_utc, site_id, ("gen", "gen-net"))

    scalar_names = {"triad-actual-gsp-kw", "triad-actual-gbp"}

    rate_names = {"triad-actual-rate", "triad-estimate-rate"}

    for i in range(1, 4):
        pref = "triad-actual-" + str(i) + "-"
        for suf in ("msp-kw", "gsp-kw"):
            scalar_names.add(pref + suf)
        for suf in ("date", "status", "laf"):
            rate_names.add(pref + suf)

    for site in sites:
        displaced_era = None
        for month_start, month_finish in sorted(
            c_months_u(start_year=year - 1, start_month=4, months=12), reverse=True
        ):
            displaced_era = chellow.e.computer.displaced_era(
                sess, caches, site, month_start, month_finish, forecast_date
            )
            if displaced_era is not None:
                break

        if displaced_era is None:
            break

        site_ds = chellow.e.computer.SiteSource(
            sess,
            site,
            march_start_utc,
            march_finish_utc,
            forecast_date,
            caches,
            displaced_era,
        )
        chellow.e.duos.duos_vb(site_ds)
        chellow.e.tnuos.hh(site_ds)

        for hh in site_ds.hh_data:
            bill_hh = site_ds.supplier_bill_hhs[hh["start-date"]]
            for k in scalar_names & hh.keys():
                bill_hh[k] = hh[k]

            for k in rate_names & hh.keys():
                bill_hh[k] = {hh[k]}

        bill = reduce_bill_hhs(site_ds.supplier_bill_hhs)
        values = [site.code, site.name]
        for i in range(1, 4):
            triad_prefix = "triad-actual-" + str(i) + "-"
            for suffix in ("date", "msp-kw", "laf", "gsp-kw"):
                values.append(csv_make_val(bill[triad_prefix + suffix]))

        for suffix in ("gsp-kw", "rate", "gbp"):
            values.append(csv_make_val(bill["triad-actual-" + suffix]))

        writer.writerow(values)

        # Avoid long-running transaction
        sess.rollback()


def content(year, site_id, user_id):
    caches = {}
    f = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("output.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            _write_sites(sess, caches, writer, year, site_id)

    except BadRequest as e:
        writer.writerow([e.description])
    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        if writer is not None:
            writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    site_id = req_int("site_id") if "site_id" in request.values else None
    year = req_int("year")
    threading.Thread(target=content, args=(year, site_id, g.user.id)).start()
    return chellow_redirect("/downloads", 303)



import re
from decimal import Decimal, InvalidOperation
from enum import Enum, auto
from io import BytesIO
from itertools import chain
from zipfile import BadZipFile, ZipFile

from openpyxl import load_workbook

from sqlalchemy import null, or_, select

from werkzeug.exceptions import BadRequest

from chellow.models import Contract, Llfc, Party, RateScript, VoltageLevel
from chellow.rate_server import download
from chellow.utils import ct_datetime, hh_format, to_utc


def get_cell(sheet, col, row):
    try:
        coordinates = f"{col}{row}"
        return sheet[coordinates]
    except IndexError:
        raise BadRequest(f"Can't find the cell {coordinates} on sheet {sheet}.")


def get_int(sheet, col, row):
    return int(get_cell(sheet, col, row).value)


def get_dec(sheet, col, row):
    cell = get_cell(sheet, col, row)
    try:
        return Decimal(str(cell.value))
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_str(sheet, col, row):
    return get_cell(sheet, col, row).value.strip()


def get_value(row, idx):
    val = row[idx].value
    if isinstance(val, str):
        return val.strip()
    else:
        return val


def get_rate(row, idx):
    cell = row[idx]
    if cell.number_format == "#":
        return None

    val = cell.value
    if val is None:
        return None
    elif isinstance(val, str) and len(val) == 0:
        return None
    else:
        try:
            return round(Decimal(val) / Decimal("100"), 5)
        except InvalidOperation as e:
            raise BadRequest(f"Can't parse the decimal '{val}' {e}.") from e


def get_rag_rate(row, idx):
    val = get_rate(row, idx)
    return get_rag_rate(row, idx - 1) if val is None else val


def get_zero_rate(row, idx):
    val = get_rate(row, idx)
    return Decimal("0.00000") if val is None else val


def get_decimal(row, idx):
    return round(Decimal(get_value(row, idx)), 5)


def to_llfcs(val):
    llfcs = []
    if isinstance(val, str):
        val = ",".join(val.splitlines())
        val = val.replace("/", ",").replace("inclusive", ",")
        for v in map(str.strip, val.split(",")):
            if len(v) > 0:
                if "-" in v:
                    start, finish = v.split("-")
                    m = re.search(r"\d", start)
                    s = m.start()
                    m = re.search(r"\d", finish)
                    f = m.start()

                    for i in range(int(start[s:]), int(finish[f:]) + 1):
                        suff = str(i).zfill(3 - s)
                        llfcs.append(f"{start[:s]}{suff}")
                else:
                    llfcs.append(v)

    elif isinstance(val, Decimal):
        llfcs.append(str(int(val)))
    elif isinstance(val, float):
        llfcs_str = str(int(val))
        for i in range(0, len(llfcs_str), 3):
            llfcs.append(llfcs_str[i : i + 3])
    elif isinstance(val, int):
        val_str = str(val)
        for i in range(0, len(val_str), 3):
            llfcs.append(val_str[i : i + 3])
    return [v.zfill(3) for v in llfcs]


def to_pcs(row, idx):
    val = get_value(row, idx)
    llfcs = []
    if isinstance(val, str):
        for v in map(str.strip, val.replace("or", ",").replace("to", "-").split(",")):
            if len(v) > 0:
                if "-" in v:
                    start, finish = v.split("-")
                    for i in range(int(start), int(finish) + 1):
                        llfcs.append(str(i))
                else:
                    llfcs.append(v)
    elif isinstance(val, Decimal):
        llfcs.append(str(int(val)))
    elif isinstance(val, float):
        llfcs_str = str(int(val))
        for i in range(0, len(llfcs_str), 3):
            llfcs.append(llfcs_str[i : i + 3])
    elif isinstance(val, int):
        val_str = str(val)
        for i in range(0, len(val_str), 3):
            llfcs.append(val_str[i : i + 3])
    return [v.zfill(2) for v in llfcs]


BAND_WEEKEND = {
    "monday to friday": False,
    "weekends": True,
    "monday to friday (including bank holidays) all year": False,
    "saturday and sunday all year": True,
    "notes": None,
}


def str_to_hr(hr_str):
    for sep in (":", "."):
        if sep in hr_str:
            break

    time_strs_raw = hr_str.strip().split(sep)
    len_time_strs_raw = len(time_strs_raw)
    if len_time_strs_raw == 1:
        ts = time_strs_raw[0]
        time_strs = ts[:2], ts[2:]
    elif len_time_strs_raw == 2:
        time_strs = time_strs_raw
    else:
        raise BadRequest(f"Can't work out the hours and minutes from '{hr_str}'")

    hours, minutes = map(Decimal, time_strs)
    return hours + minutes / Decimal(60)


def val_to_slots(val):
    slots = []
    time_str = None if val is None else str(val).strip()
    if time_str not in (None, "", "[Start] - [End]", "0"):
        for t_str in time_str.splitlines():
            for sep in ("-", "to"):
                if sep in t_str:
                    break
            start_str, finish_str = t_str.split(sep)
            slots.append(
                {"start": str_to_hr(start_str), "finish": str_to_hr(finish_str)}
            )
    return slots


def col_find(row, pattern, repeats=1):
    for i, cell in enumerate(row):
        txt = cell.value
        if txt is not None:
            txt_str = " ".join(str(txt).lower().split())
            if re.search(pattern, txt_str) is not None:
                if repeats == 1:
                    return i
                else:
                    repeats -= 1


def col_match(row, pattern, repeats=1):
    result = col_find(row, pattern, repeats=repeats)
    if result is None:
        raise BadRequest(
            f"Pattern '{pattern}' not found in row "
            + ", ".join(str(cell.value) for cell in row)
        )
    return result


def tab_lv_hv(sheet, gsp_rates):
    try:
        tariffs = gsp_rates["tariffs"]
    except KeyError:
        tariffs = gsp_rates["tariffs"] = {}

    bands = gsp_rates["bands"] = []

    class State(Enum):
        OUTSIDE = auto()
        BANDS = auto()
        TARIFFS = auto()

    state = State.OUTSIDE
    title_row = None
    for row in sheet.iter_rows():
        val = get_value(row, 0)
        val_0 = None if val is None else " ".join(val.split())
        val_0_lower = None if val is None else val_0.lower()
        if state == State.OUTSIDE:
            if val_0_lower == "tariff name" or get_value(row, 1) == "Open LLFCs":
                state = State.TARIFFS
                title_row = row
            elif val_0_lower == "time periods":
                state = State.BANDS

        elif state == State.BANDS:
            try:
                is_weekend = BAND_WEEKEND[val_0_lower]
            except KeyError:
                raise BadRequest(f"The key '{val_0_lower}' isn't recognised.")

            if is_weekend is None:
                state = State.OUTSIDE
            else:
                for i, band_name in enumerate(("red", "amber")):
                    for slot in val_to_slots(get_value(row, i + 1)):
                        bands.append(
                            {
                                "weekend": is_weekend,
                                "start": slot["start"],
                                "finish": slot["finish"],
                                "band": band_name,
                            }
                        )

        elif state == State.TARIFFS:
            if val_0 is None or len(val_0) == 0:
                state = State.OUTSIDE
            else:
                llfcs_str = ",".join(
                    chain(to_llfcs(get_value(row, 1)), to_llfcs(get_value(row, 10)))
                )
                pcs_str = ",".join(to_pcs(row, 2))
                tariffs[llfcs_str + "_" + pcs_str] = {
                    "description": val_0,
                    "gbp-per-mpan-per-day": get_zero_rate(
                        row, col_match(title_row, "fixed")
                    ),
                    "gbp-per-kva-per-day": get_zero_rate(
                        row, col_match(title_row, "^capacity")
                    ),
                    "excess-gbp-per-kva-per-day": get_zero_rate(
                        row, col_match(title_row, "exce")
                    ),
                    "red-gbp-per-kwh": get_rag_rate(row, col_match(title_row, "red")),
                    "amber-gbp-per-kwh": get_rag_rate(
                        row, col_match(title_row, "amber")
                    ),
                    "green-gbp-per-kwh": get_rag_rate(
                        row, col_match(title_row, "green")
                    ),
                    "gbp-per-kvarh": get_zero_rate(
                        row, col_match(title_row, "reactive")
                    ),
                }


# State for EHV

EHV_BLANK = 0
EHV_BANDS = 1
EHV_TARIFFS = 2


def tab_ehv(sheet, gsp_rates):
    try:
        tariffs = gsp_rates["tariffs"]
    except KeyError:
        tariffs = gsp_rates["tariffs"] = {}

    bands = gsp_rates["super_red"] = []

    state = EHV_BLANK
    title_row = None
    for row in sheet.iter_rows():
        val = get_value(row, 0)
        val_0 = None if val is None else " ".join(str(val).split()).lower()
        if state == EHV_BLANK:
            if val_0 == "time periods":
                state = EHV_BANDS
                title_row = row
            elif val_0 in ("import unique identifier", "import llfc"):
                state = EHV_TARIFFS
                title_row = row

        elif state == EHV_TARIFFS:
            for polarity, repeats in (("import", 1), ("export", 2)):
                llfc_val = get_value(row, col_match(title_row, "llfc", repeats=repeats))
                llfc = None if llfc_val is None else str(llfc_val).strip()
                if llfc not in (None, ""):
                    band_col = col_find(title_row, "residual")
                    band = "" if band_col is None else int(get_decimal(row, band_col))
                    tariffs[llfc] = {
                        "super-red-gbp-per-kwh": get_rate(
                            row, col_match(title_row, polarity + " super red")
                        ),
                        "gbp-per-day": get_zero_rate(
                            row, col_match(title_row, polarity + " fixed")
                        ),
                        "gbp-per-kva-per-day": get_zero_rate(
                            row, col_match(title_row, polarity + " capacity")
                        ),
                        "excess-gbp-per-kva-per-day": get_zero_rate(
                            row, col_match(title_row, polarity + " exce")
                        ),
                        "description": f"Designated EHV{band}",
                    }

        elif state == EHV_BANDS:
            if val_0 in (None, "", "notes"):
                state = EHV_BLANK
            else:
                period_str = " ".join(get_value(row, 0).split()).lower()
                periods = []

                if period_str in (
                    "monday to friday nov to feb (excluding "
                    "22nd dec to 4th jan inclusive)",
                    "monday to friday (including bank holidays) nov to feb inclusive "
                    "(excluding 22nd dec to 4th jan inclusive)",
                ):
                    periods.append(
                        {
                            "weekend": False,
                            "start-month": 11,
                            "start-day": 1,
                            "finish-month": 12,
                            "finish-day": 21,
                        }
                    )
                    periods.append(
                        {
                            "weekend": False,
                            "start-month": 1,
                            "start-day": 5,
                            "finish-month": 2,
                            "finish-day": "last",
                        }
                    )

                elif period_str in (
                    "monday to friday (including bank holidays) november to february",
                    "monday to friday nov to feb",
                ):
                    periods.append(
                        {
                            "weekend": False,
                            "start-month": 11,
                            "start-day": 1,
                            "finish-month": 2,
                            "finish-day": "last",
                        }
                    )

                for slot in val_to_slots(get_value(row, 3)):
                    for period in periods:
                        bands.append(
                            {
                                **period,
                                "start-hour": slot["start"],
                                "finish-hour": slot["finish"],
                            }
                        )


def find_rates(file_name, file_like):
    rates = {"a_file_name": file_name}
    vls = []

    if file_name.endswith(".zip"):
        try:
            with ZipFile(file_like) as za:
                for zname in za.namelist():
                    if zname.endswith(".xlsx"):
                        with za.open(zname) as f:
                            gsp_code, gsp_rates, gsp_vls = find_gsp_group_rates(
                                zname, BytesIO(f.read())
                            )
                            rates[gsp_code] = gsp_rates
                            vls.extend(gsp_vls)
        except BadZipFile as e:
            raise BadRequest(f"Problem with zip file '{file_name}'") from e

    elif file_name.endswith(".xlsx"):
        gsp_code, gsp_rates, gsp_vls = find_gsp_group_rates(file_name, file_like)
        rates[gsp_code] = gsp_rates
        vls.extend(gsp_vls)

    else:
        raise BadRequest(f"The file extension for {file_name} isn't recognized.")

    return rates, vls


GSP_MAP = (
    ("gsp a", "_A"),
    ("gsp b", "_B"),
    ("gsp c", "_C"),
    ("gsp d", "_D"),
    ("gsp e", "_E"),
    ("gsp f", "_F"),
    ("gsp g", "_G"),
    ("gsp h", "_H"),
    ("gsp j", "_J"),
    ("gsp k", "_K"),
    ("gsp l", "_L"),
    ("gsp m", "_M"),
    ("gsp n", "_N"),
    ("gsp p", "_P"),
    ("gsp_a", "_A"),
    ("gsp_b", "_B"),
    ("gsp_c", "_C"),
    ("gsp_d", "_D"),
    ("gsp_e", "_E"),
    ("gsp_f", "_F"),
    ("gsp_g", "_G"),
    ("gsp_h", "_H"),
    ("gsp_j", "_J"),
    ("gsp_k", "_K"),
    ("gsp_l", "_L"),
    ("gsp_m", "_M"),
    ("gsp_n", "_N"),
    ("gsp_p", "_P"),
    ("lond", "_C"),
    ("mide", "_E"),
    ("sepd", "_H"),
    ("sweb", "_L"),
)


def find_gsp_group_rates(file_name, file_like):
    rates = {}

    fname = file_name.lower()
    gsp_code = None
    for substr, code in GSP_MAP:
        if substr in fname:
            gsp_code = code
            break

    if gsp_code is None:
        raise BadRequest(
            f"Can't determine the GSP group from the file name '{file_name}'."
        )

    book = load_workbook(file_like, data_only=True)

    vls = None

    try:
        for sheet in book.worksheets:
            title = sheet.title.strip().lower()
            if title.startswith("annex 1 "):
                tab_lv_hv(sheet, rates)
            elif title.startswith("annex 2 "):
                tab_ehv(sheet, rates)
            elif title.startswith("annex 5 "):
                vls = tab_llfs(sheet)
    except BadRequest as e:
        raise BadRequest(f"Problem with file '{file_name}': {e.description}")

    return gsp_code, rates, vls


def rate_server_import(sess, log, set_progress, s, paths):
    log("Starting to check for new DNO spreadsheets")
    year_entries = {}
    for path, url in paths:
        if len(path) == 5:
            year_str, utility, rate_type, dno_code, file_name = path
            year = int(year_str)
            try:
                dno_entries = year_entries[year]
            except KeyError:
                dno_entries = year_entries[year] = {}

            if utility == "electricity" and rate_type == "duos":
                try:
                    fl_entries = dno_entries[dno_code]
                except KeyError:
                    fl_entries = dno_entries[dno_code] = {}

                fl_entries[file_name] = url

    for year, dno_entries in sorted(year_entries.items()):
        fy_start = to_utc(ct_datetime(year, 4, 1))
        for dno_code, fl_entries in sorted(dno_entries.items()):
            contract = Contract.get_dno_by_name(sess, dno_code)
            if fy_start < contract.start_rate_script.start_date:
                continue

            rs = sess.execute(
                select(RateScript).where(
                    RateScript.contract == contract,
                    RateScript.start_date == fy_start,
                )
            ).scalar_one_or_none()
            if rs is None:
                rs = contract.insert_rate_script(sess, fy_start, {})

            file_name, url = sorted(fl_entries.items())[-1]

            rs_script = rs.make_script()
            if rs_script.get("a_file_name") != file_name:
                try:
                    fl = BytesIO(download(s, url))
                    rates, vls = find_rates(file_name, fl)
                    rs.update(rates)
                    log(
                        f"Updated DNO {dno_code} rate script for "
                        f"{hh_format(fy_start)}"
                    )
                    update_vls(sess, log, vls, dno_code, fy_start, rs.finish_date)
                except BadRequest as e:
                    raise BadRequest(
                        f"Problem with year {year} DNO {dno_code} file name "
                        f"{file_name}: {e.description}"
                    )
                except BaseException as e:
                    raise BadRequest(
                        f"Problem with year {year} DNO {dno_code} file name "
                        f"{file_name}:"
                    ) from e

    log("Finished DNO spreadsheets")
    sess.commit()


LV_NET = ("LV", False)
LV_SUB = ("LV", True)
HV_NET = ("HV", False)
HV_SUB = ("HV", True)
EHV_NET = ("EHV", False)
EHV_SUB = ("EHV", True)


VL_LOOKUP = {
    "132/33kv": EHV_NET,
    "132/33kv generic": EHV_NET,
    "132/33kv substation": EHV_SUB,
    "132/ehv connected": EHV_NET,
    "132/hv connected": HV_NET,
    "132kv connected": EHV_NET,
    "132kv generic": EHV_NET,
    "132kv generic (demand)": EHV_NET,
    "132kv generic (generation)": EHV_NET,
    "132kv generic export": EHV_NET,
    "132kv generic import": EHV_NET,
    "132kv to 33k generic": EHV_NET,
    "132kv to 33kv generic": EHV_NET,
    "33kv generic": EHV_NET,
    "33kv generic (demand)": EHV_NET,
    "33kv generic (generation)": EHV_NET,
    "33kv generic export": EHV_NET,
    "33kv generic import": EHV_NET,
    "ehv 33kv export": EHV_NET,
    "ehv 33kv import": EHV_NET,
    "ehv connected": EHV_NET,
    "greater than 22kv connected - demand": HV_NET,
    "greater than 22kv connected - generation": HV_NET,
    "high voltage network": HV_NET,
    "high voltage network export": HV_NET,
    "high voltage network import": HV_NET,
    "high-voltage network": HV_NET,
    "high voltage substation": HV_SUB,
    "high voltage substation export": HV_SUB,
    "high voltage substation import": HV_SUB,
    "high-voltage substation": HV_SUB,
    "low voltage network": LV_NET,
    "low voltage network export": LV_NET,
    "low voltage network import": LV_NET,
    "low-voltage network": LV_NET,
    "low voltage substation": LV_SUB,
    "low voltage substation export": LV_SUB,
    "low voltage substation import": LV_SUB,
    "low-voltage substation": LV_SUB,
}


def tab_llfs(sheet):
    llfcs = []

    in_llfcs = False
    for row in range(1, len(sheet["A"]) + 1):
        val = get_cell(sheet, "A", row).value
        val_0 = None if val is None else " ".join(val.split())
        if in_llfcs:
            if val_0 is None or len(val_0) == 0:
                in_llfcs = False
            else:
                vl, is_substation = VL_LOOKUP[val_0.lower()]
                for llfc_code in to_llfcs(get_cell(sheet, "H", row).value):
                    llfc = {
                        "code": llfc_code,
                        "voltage_level": vl,
                        "is_substation": is_substation,
                    }
                    llfcs.append(llfc)

        elif val_0 == "Metered voltage":
            in_llfcs = True

    return llfcs


def update_vls(sess, logger, vls, dno_code, fy_start, rs_finish):
    dno = Party.get_dno_by_code(sess, dno_code, fy_start)

    for vl in vls:
        vl_code = vl["code"]
        q = select(Llfc).where(
            Llfc.dno == dno,
            Llfc.code == vl_code,
            or_(Llfc.valid_to == null(), Llfc.valid_to >= fy_start),
        )
        if rs_finish is not None:
            q = q.where(Llfc.valid_from <= rs_finish)
        llfc = sess.execute(q).scalar_one_or_none()

        if llfc is None:
            raise BadRequest(
                f"There is no LLFC with the code '{vl_code}' associated with the DNO "
                f"{dno.code} from {hh_format(fy_start)} to {hh_format(rs_finish)}."
            )

        vl_voltage_level = VoltageLevel.get_by_code(sess, vl["voltage_level"])
        llfc.voltage_level = vl_voltage_level

        llfc.is_substation = vl["is_substation"]
        if sess.is_modified(llfc):
            logger(f"Updated LLFC {llfc.code} of DNO {dno_code}")
            sess.flush()



import csv
import sys
import threading
import traceback

from flask import g, request

from sqlalchemy import or_
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.e.computer import contract_func
from chellow.gas.engine import GDataSource, forecast_date
from chellow.models import GEra, GSupply, Session
from chellow.utils import csv_make_val, req_date, req_int
from chellow.views import chellow_redirect


def content(date, g_supply_id, user):
    try:
        with Session() as sess:
            f = open_file("g_supplies_snapshot.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            writer.writerow(
                (
                    "Date",
                    "Physical Site Id",
                    "Physical Site Name",
                    "Other Site Ids",
                    "Other Site Names",
                    "MPRN",
                    "Exit Zone",
                    "Meter Serial Number",
                    "Correction Factor",
                    "Unit",
                    "Contract",
                    "Account",
                    "Supply Start",
                    "Supply Finish",
                    "reading_frequency",
                    "aq",
                    "soq",
                )
            )
            fdate = forecast_date()
            caches = {}

            g_eras = (
                sess.query(GEra, GSupply)
                .join(GSupply)
                .filter(
                    GEra.start_date <= date,
                    or_(GEra.finish_date == null(), GEra.finish_date >= date),
                )
                .order_by(GSupply.mprn)
                .options(joinedload(GEra.site_g_eras))
            )

            if g_supply_id is not None:
                g_supply = GSupply.get_by_id(sess, g_supply_id)

                g_eras = g_eras.filter(GEra.g_supply == g_supply)

            for g_era, g_supply in g_eras:
                site_codes = []
                site_names = []
                for site_g_era in g_era.site_g_eras:
                    if site_g_era.is_physical:
                        physical_site = site_g_era.site
                    else:
                        site = site_g_era.site
                        site_codes.append(site.code)
                        site_names.append(site.name)

                sup_g_eras = (
                    sess.query(GEra)
                    .filter(GEra.g_supply == g_supply)
                    .order_by(GEra.start_date)
                    .all()
                )
                g_supply_start_date = sup_g_eras[0].start_date
                g_supply_finish_date = sup_g_eras[-1].finish_date

                ds = GDataSource(sess, date, date, fdate, g_era, caches, None)
                g_supplier_contract = g_era.g_contract
                vb_function = contract_func(caches, g_supplier_contract, "virtual_bill")
                if vb_function is None:
                    raise BadRequest(
                        f"The contract {g_supplier_contract.name} doesn't have the "
                        f"virtual_bill() function."
                    )
                vb_function(ds)
                bill = ds.bill

                writer.writerow(
                    csv_make_val(value)
                    for value in [
                        date,
                        physical_site.code,
                        physical_site.name,
                        site_codes,
                        site_names,
                        g_supply.mprn,
                        g_supply.g_exit_zone.code,
                        g_era.msn,
                        g_era.correction_factor,
                        g_era.g_unit.code,
                        g_supplier_contract.name,
                        g_era.account,
                        g_supply_start_date,
                        g_supply_finish_date,
                        g_era.g_reading_frequency.code,
                        bill["aq"],
                        bill["soq"],
                    ]
                )
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(session):
    user = g.user
    date = req_date("date")
    if "g_supply_id" in request.values:
        g_supply_id = req_int("g_supply_id")
    else:
        g_supply_id = None

    args = (date, g_supply_id, user)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from io import StringIO

from utils import match

from chellow.models import User
from chellow.reports.report_batches import content


def test_do_get(client, mocker, sess):
    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_batches.open_file", return_value=mock_file)
    MockThread = mocker.patch("chellow.reports.report_batches.threading.Thread")
    user = User.get_by_email_address(sess, "admin@example.com")
    expected_args = (user.id,)
    response = client.get("/reports/batches")
    match(response, 303)
    MockThread.assert_called_with(target=content, args=expected_args)



from pathlib import Path
from tempfile import TemporaryDirectory

from chellow.dloads import open_file, startup


def test_startup(mocker):
    mocker.patch("chellow.dloads.FileDeleter")
    with TemporaryDirectory() as tmpdir:
        startup(Path(tmpdir))


def test_truncate(mocker):
    mocker.patch("chellow.dloads.FileDeleter")
    with TemporaryDirectory() as tmpdir:
        startup(Path(tmpdir))
        user = None
        with open_file("test.txt", user, mode="w") as fl:
            fl.truncate()



import csv
import sys
import threading
import traceback
from datetime import datetime as Datetime

from flask import g, request

from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true

import chellow.e.tnuos
from chellow.dloads import open_file
from chellow.e.computer import SupplySource, forecast_date
from chellow.e.duos import duos_vb
from chellow.models import Era, Pc, Session, Site, SiteEra, Source, Supply, User
from chellow.utils import (
    HH,
    csv_make_val,
    ct_datetime,
    hh_format,
    reduce_bill_hhs,
    req_int,
    to_utc,
)
from chellow.views import chellow_redirect


def _make_eras(sess, nov_start, year_finish, supply_id):
    eras = (
        sess.query(Era)
        .join(Supply)
        .join(Source)
        .join(Pc)
        .filter(
            Era.start_date <= year_finish,
            or_(Era.finish_date == null(), Era.finish_date >= nov_start),
            Source.code.in_(("net", "gen-net")),
            Pc.code == "00",
        )
        .order_by(Supply.id)
    )

    if supply_id is not None:
        eras = eras.filter(Supply.id == supply_id)

    return eras


def content(year, supply_id, user_id):
    caches = {}
    f = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("supplies_triad.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")

            march_start = to_utc(ct_datetime(year, 3, 1))
            march_finish = to_utc(ct_datetime(year, 4, 1)) - HH
            nov_start = to_utc(ct_datetime(year - 1, 11, 1))

            scalar_names = {
                "triad-actual-gsp-kw",
                "triad-actual-gbp",
                "triad-estimate-gsp-kw",
                "triad-estimate-months",
                "triad-estimate-gbp",
                "triad-all-estimates-months",
                "triad-all-estimates-gbp",
            }

            rate_names = {"triad-actual-rate", "triad-estimate-rate"}

            for i in range(1, 4):
                for p in ("triad-actual-", "triad-estimate-"):
                    act_pref = f"{p}{i}-"
                    for suf in ("msp-kw", "gsp-kw"):
                        scalar_names.add(act_pref + suf)
                    for suf in ("date", "status", "laf"):
                        rate_names.add(act_pref + suf)

            def triad_csv(supply_source):
                if supply_source is None or supply_source.mpan_core.startswith("99"):
                    return [""] * 19

                duos_vb(supply_source)
                chellow.e.tnuos.hh(supply_source)
                for hh in supply_source.hh_data:
                    bill_hh = supply_source.supplier_bill_hhs[hh["start-date"]]
                    for k in scalar_names & hh.keys():
                        bill_hh[k] = hh[k]

                    for k in rate_names & hh.keys():
                        bill_hh[k] = {hh[k]}
                bill = reduce_bill_hhs(supply_source.supplier_bill_hhs)
                values = [supply_source.mpan_core]
                for i in range(1, 4):
                    triad_prefix = "triad-actual-" + str(i)
                    for suffix in ["-date", "-msp-kw", "-status", "-laf", "-gsp-kw"]:
                        values.append(csv_make_val(bill[triad_prefix + suffix]))

                suffixes = ["gsp-kw", "rate", "gbp"]
                values += [
                    csv_make_val(bill["triad-actual-" + suf]) for suf in suffixes
                ]
                return values

            writer.writerow(
                (
                    "Site Code",
                    "Site Name",
                    "Supply Name",
                    "Source",
                    "Generator Type",
                    "Import MPAN Core",
                    "Import T1 Date",
                    "Import T1 MSP kW",
                    "Import T1 Status",
                    "Import T1 LAF",
                    "Import T1 GSP kW",
                    "Import T2 Date",
                    "Import T2 MSP kW",
                    "Import T2 Status",
                    "Import T2 LAF",
                    "Import T2 GSP kW",
                    "Import T3 Date",
                    "Import T3 MSP kW",
                    "Import T3 Status",
                    "Import T3 LAF",
                    "Import T3 GSP kW",
                    "Import GSP kW",
                    "Import Rate GBP / kW",
                    "Import GBP",
                    "Export MPAN Core",
                    "Export T1 Date",
                    "Export T1 MSP kW",
                    "Export T1 Status",
                    "Export T1 LAF",
                    "Export T1 GSP kW",
                    "Export T2 Date",
                    "Export T2 MSP kW",
                    "Export T2 Status",
                    "Export T2 LAF",
                    "Export T2 GSP kW",
                    "Export T3 Date",
                    "Export T3 MSP kW",
                    "Export T3 Status",
                    "Export T3 LAF",
                    "Export T3 GSP kW",
                    "Export GSP kW",
                    "Export Rate GBP / kW",
                    "Export GBP",
                )
            )

            fdate = forecast_date()
            eras = _make_eras(sess, nov_start, march_finish, supply_id)

            for era in eras:
                site = (
                    sess.query(Site)
                    .join(SiteEra)
                    .filter(SiteEra.is_physical == true(), SiteEra.era == era)
                    .one()
                )
                supply = era.supply

                imp_mpan_core = era.imp_mpan_core
                if imp_mpan_core is None:
                    imp_supply_source = None
                else:
                    imp_supply_source = SupplySource(
                        sess, march_start, march_finish, fdate, era, True, caches
                    )

                exp_mpan_core = era.exp_mpan_core
                if exp_mpan_core is None:
                    exp_supply_source = None
                else:
                    exp_supply_source = SupplySource(
                        sess, march_start, march_finish, fdate, era, False, caches
                    )

                gen_type = supply.generator_type
                gen_type = "" if gen_type is None else gen_type.code
                vals = []
                for value in (
                    [site.code, site.name, supply.name, supply.source.code, gen_type]
                    + triad_csv(imp_supply_source)
                    + triad_csv(exp_supply_source)
                ):
                    if isinstance(value, Datetime):
                        vals.append(hh_format(value))
                    else:
                        vals.append(str(value))
                writer.writerow(vals)

                # Avoid a long-running transaction
                sess.rollback()
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    year = req_int("year")
    supply_id = req_int("supply_id") if "supply_id" in request.values else None
    threading.Thread(target=content, args=(year, supply_id, g.user.id)).start()
    return chellow_redirect("/downloads", 303)



from decimal import Decimal
from io import StringIO

from chellow.models import (
    BillType,
    Comm,
    Contract,
    Cop,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    ReadType,
    Site,
    Source,
    Tpr,
    User,
    UserRole,
    VoltageLevel,
    insert_bill_types,
    insert_comms,
    insert_cops,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_219 import content
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_content(mocker, sess):
    valid_from = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", valid_from, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", valid_from, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", valid_from, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", valid_from, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(sess, "845", False, True, valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    batch = imp_supplier_contract.insert_batch(sess, "a", "batch")

    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")
    bill = batch.insert_bill(
        sess,
        "xxx",
        "543",
        to_utc(ct_datetime(2020, 2, 3)),
        to_utc(ct_datetime(2020, 1, 3)),
        to_utc(ct_datetime(2020, 1, 6)),
        Decimal("0"),
        Decimal("0.00"),
        Decimal("0.00"),
        Decimal("0.00"),
        bill_type,
        {},
        supply,
    )
    tpr = Tpr.insert(sess, "00001", False, True)
    read_type = ReadType.insert(sess, "A", "Actual")
    bill.insert_read(
        sess,
        tpr,
        Decimal(1),
        "kWh",
        "5457339",
        "3782897",
        to_utc(ct_datetime(2020, 1, 3)),
        Decimal("66"),
        read_type,
        to_utc(ct_datetime(2020, 1, 5)),
        Decimal("100"),
        read_type,
    )

    sess.commit()
    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_219.open_file", return_value=mock_file)
    year = 2020
    month = 1
    months = 1
    supply_id = supply.id
    content(year, month, months, supply_id, user.id)
    actual = mock_file.getvalue()
    expected_lines = [
        [
            "Duration Start",
            "Duration Finish",
            "Supply Id",
            "Import MPAN Core",
            "Export MPAN Core",
            "Batch Reference",
            "Bill Id",
            "Bill Reference",
            "Bill Issue Date",
            "Bill Type",
            "Register Read Id",
            "TPR",
            "Coefficient",
            "Previous Read Date",
            "Previous Read Value",
            "Previous Read Type",
            "Present Read Date",
            "Present Read Value",
            "Present Read Type",
        ],
        [
            "2020-01-01 00:00",
            "2020-01-31 23:30",
            "1",
            "22 7867 6232 781",
            "",
            "a",
            "1",
            "543",
            "2020-02-03 00:00",
            "N",
            "1",
            "00001",
            "1",
            "2020-01-03 00:00",
            "66",
            "A",
            "2020-01-05 00:00",
            "100",
            "A",
        ],
    ]
    expected = "\n".join(",".join(line) for line in expected_lines) + "\n"
    assert actual == expected



from datetime import datetime as Datetime
from io import BytesIO

from openpyxl import Workbook

from chellow.e.bill_parsers.activity_mop_stark_xlsx import Parser


def test_blank():
    f = BytesIO()
    wb = Workbook()
    wb.create_sheet("Summary")
    wb.create_sheet("NHH")
    wb.save(f)
    f.seek(0)
    parser = Parser(f)
    assert parser.book.data_only


def test_no_bills(sess):
    f = BytesIO()
    wb = Workbook()
    wb.create_sheet("NHH")
    sheet = wb.worksheets[1]
    sheet.insert_rows(0, 10)
    sheet.insert_cols(0, 10)
    sheet["B11"].value = "MPAN"
    sheet["C6"].value = Datetime(2022, 3, 1)
    wb.save(f)
    f.seek(0)
    p = Parser(f)
    p.make_raw_bills()



import csv
from datetime import datetime as Datetime, timedelta as Timedelta
from decimal import Decimal
from io import BytesIO, StringIO
from zipfile import ZipFile

# from sqlalchemy.dialects.postgresql import insert
from sqlalchemy import text

from werkzeug.exceptions import BadRequest

from chellow.models import Contract, Party
from chellow.rate_server import download
from chellow.utils import hh_after, to_ct, to_utc


def _process(sess, log, set_progress, file_like):
    zip_file = ZipFile(file_like)
    name_list = zip_file.namelist()
    if len(name_list) != 1:
        raise BadRequest("The zip archive must contain exactly one file.")
    stmt = text(
        """
INSERT INTO laf (llfc_id, timestamp, value) VALUES
(unnest(CAST(:llfc_ids AS INTEGER[])), unnest(CAST(:timestamps AS TIMESTAMPTZ[])),
unnest(CAST(:values AS NUMERIC[])))
ON CONFLICT ON CONSTRAINT laf_llfc_id_timestamp_key
DO UPDATE SET (llfc_id, timestamp, value) =
(EXCLUDED.llfc_id, EXCLUDED.timestamp, EXCLUDED.value)"""
    )
    fname = name_list[0]
    csv_file = StringIO(zip_file.read(fname).decode("utf-8"))
    csv_dt = to_utc(to_ct(Datetime.strptime(fname[-12:-4], "%Y%m%d")))
    for llfc_ids, timestamps, values in laf_days(
        sess, log, set_progress, csv_file, csv_dt
    ):
        sess.execute(
            stmt,
            params={"llfc_ids": llfc_ids, "timestamps": timestamps, "values": values},
        )
        sess.commit()


UTC_DATETIME_MIN = to_utc(Datetime.min)


def laf_days(sess, log, set_progress, csv_file, csv_dt):
    llfc_ids = []
    timestamps = []
    values = []
    llfc_code = line_dt_ct = dno = llfc_id = None
    llfc_valid_to = UTC_DATETIME_MIN
    timestamp_cache = {}

    for line_number, vals in enumerate(csv.reader(csv_file, delimiter="|")):
        set_progress(f"Reached line number {line_number}")
        code = vals[0]

        if code == "DIS":
            participant_code = vals[1]
            dno = Party.get_by_participant_code_role_code(
                sess, participant_code, "R", csv_dt
            )

        elif code == "LLF":
            llfc_code = vals[1]
            llfc_valid_to = UTC_DATETIME_MIN

            if len(llfc_ids) > 0:
                yield llfc_ids, timestamps, values
                llfc_ids = []
                timestamps = []
                values = []

        elif code == "SDT":
            line_dt_str = vals[1]
            line_dt_ct = to_ct(Datetime.strptime(line_dt_str, "%Y%m%d"))

        elif code == "SPL":
            period, value = vals[1:]

            try:
                timestamp = timestamp_cache[line_dt_ct][period]
            except KeyError:
                try:
                    day_cache = timestamp_cache[line_dt_ct]
                except KeyError:
                    day_cache = timestamp_cache[line_dt_ct] = {}

                timestamp = day_cache[period] = to_utc(
                    line_dt_ct + Timedelta(minutes=30 * (int(period) - 1))
                )

            if hh_after(timestamp, llfc_valid_to):
                llfc = dno.find_llfc_by_code(sess, llfc_code, timestamp)
                if llfc is None:
                    continue

                llfc_id, llfc_valid_to = llfc.id, llfc.valid_to
            llfc_ids.append(llfc_id)
            timestamps.append(timestamp)
            values.append(Decimal(value))

        elif code == "ZPT":
            if len(llfc_ids) > 0:
                yield llfc_ids, timestamps, values


LAF_START = "llf"
LAF_END = "ptf.zip"


def find_participant_entries(paths, laf_state):
    participant_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year_str, utility, rate_type, file_name = path

            if utility == "electricity" and rate_type == "lafs":
                # File name llfetcl20220401ptf.zip
                if not file_name.startswith(LAF_START):
                    raise BadRequest(f"A laf file must begin with '{LAF_START}'")
                if not file_name.endswith(LAF_END):
                    raise BadRequest(f"A laf file must end with '{LAF_END}'")

                participant_code = file_name[3:7]
                timestamp = file_name[7:15]

                participant_state = laf_state.get(participant_code, "")
                if timestamp > participant_state:
                    try:
                        fl_entries = participant_entries[participant_code]
                    except KeyError:
                        fl_entries = participant_entries[participant_code] = {}
                    fl_entries[timestamp] = url
    return participant_entries


def rate_server_import(sess, log, set_progress, s, paths):
    log("Starting to check for new LAF files")
    conf = Contract.get_non_core_by_name(sess, "configuration")
    state = conf.make_state()

    try:
        laf_state = state["laf_importer"]
    except KeyError:
        laf_state = state["laf_importer"] = {}

    participant_entries = find_participant_entries(paths, laf_state)

    for participant_code, fl_entries in sorted(participant_entries.items()):
        log(f"Importing files for participant code {participant_code}")
        for timestamp, url in sorted(fl_entries.items()):
            log(f"Importing file with timestamp {timestamp}")

            fl = BytesIO(download(s, url))
            _process(sess, log, set_progress, fl)
            laf_state[participant_code] = timestamp
            conf.update_state(state)
            sess.commit()

    log("Finished LAF files")
    sess.commit()
